(*-----------#pc-version-----------------------------------------------------*)
(*
implementation
uses io,util,rsutil,calctet,calctet1,sys;
*)
(*-----------#pc-version-----------------------------------------------------*)

(*-----------#unix-version-----------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "mon.h"
#include "mon1.h"
#include "calctet.h"
#include "calctet1.h"
#include "io.h"
#include "sys.h"
#include "util.h"
#include "eqs.h"
#include "scal.h"
#include "rsutil.h"
#include "cspecial.h"
#include "makemod.h"
#include "cross.h"
#include "linbayes.h"
#include "gib.h"
#include "up.h"

type
   whererec = record
                varnum: vertextype;
                val: categorynum;
              end;
 
   wherearray = array[1..maxvertex] of whererec;

var theparents,thechildren,cluster,oldcluster: tempset;
    where: wherearray;


procedure initcount(curpos: probptr; v: vertextype);
var cut: real; i: vertextype; numinloop: real;
begin
   if curpos <> nil then
    if curpos^.nextvalue[1] <> nil then
      begin
        i := 1;
        while curpos^.nextvalue[i] <> nil do 
          begin
            initcount(curpos^.nextvalue[i],i);
            i := i + 1;
          end;
      end
    else for i := 1 to maxval[curpos^.varnum] do curpos^.count[i] := 0;
end;

procedure initlatentparms(curpos: probptr; v: vertextype);
var cut: real; i: vertextype; numinloop: real;
begin
   if curpos <> nil then
    if curpos^.nextvalue[1] <> nil then
      begin
        i := 1;
        while curpos^.nextvalue[i] <> nil do 
          begin
            initlatentparms(curpos^.nextvalue[i],i);
            i := i + 1;
          end;
      end
    else 
      begin
        curpos^.cutoff[0] := 0;
        curpos^.cutoff[maxval[v]] := 1.0;
        numinloop := randunif(0.01,0.99);
        for i := 1 to maxval[v]-1 do 
           begin
              curpos^.cutoff[i] := curpos^.cutoff[i-1] + numinloop;
              numinloop := (1 - numinloop) * 0.5;
           end;
        curpos^.multiplier[1] := -1;
      end;
end;



procedure findlatentpars(var adjacency: adjtype; var haslatentpar: vertexset);
var i,j: vertextype;
begin
   haslatentpar := [numob+1..max];
   for i := 1 to numob do
      for j := numob + 1 to max do
        if adjacency^[j,i] <> notthere then haslatentpar := haslatentpar + [i];
end;


procedure initlatents(var adjacency: adjtype);
var i,j: vertextype; newseed: integer;
begin
   initrand(newseed);
   findlatentpars(adjacency,haslatentpar);
   for i := numob + 1 to max do
      begin
        maxval[i] := 2;
        cats[i].num := 2;
        for j := 1 to maxval[i] do varcatvalue[i,j] := j;
      end;
   formtree(adjacency,cats);
   for i := 1 to max do if i in haslatentpar then initlatentparms(cats[i].condprob,i);
end;

procedure getparents(var adjacency: adjtype; v: vertextype; var parents: vertexset);
var i: vertextype; temp: vertexset;
begin
  temp := [];
  for i := 1 to max do 
    if adjacency^[i,v] <> notthere then 
      temp := temp + [i];
  parents := temp;
end;

procedure getchildren(var adjacency: adjtype; v: vertextype; var children: vertexset);
var i: vertextype; temp: vertexset;
begin
  temp := [];
  for i := 1 to max do 
    if adjacency^[v,i] <> notthere then 
      temp := temp + [i];
  children := temp;
end;

procedure getlatentclust(var adjacency: adjtype);
var i,j,curpos,last: vertextype; done: vertexset; localadj: adjtype;
begin
  tnewadj(localadj,fatal);
  for i := 1 to max do
    begin
      getparents(adjacency,i,theparents[i]); 
      getchildren(adjacency,i,thechildren[i]);   
    end;  
  done := [];
  curpos := 1;
  for i := numob + 1 to max do
    begin
      if (not (i in done)) then
        begin
           cluster[curpos] := [i] + thechildren[i];   
           for j := 1 to max do               
              if (j in cluster[curpos]) then cluster[curpos] := cluster[curpos] + theparents[j];
           repeat
              oldcluster[curpos] := cluster[curpos];            
              for j := i+1 to max do
                 if (j in cluster[curpos]) then cluster[curpos] := cluster[curpos] + thechildren[j];
              for j := 1 to max do
                 if (j in cluster[curpos]) then cluster[curpos] := cluster[curpos] + theparents[j];
           until oldcluster[curpos] = cluster[curpos];
           done := done + cluster[curpos];
           curpos := curpos + 1;
        end;
     end;
  curpos := 1;
  for i := 1 to max do tempcats[i].num := maxval[i];
  while cluster[curpos] <> [] do
     begin
       for i := 1 to max do 
         begin
             if (i in cluster[curpos]) then last := i;
             for j := 1 to max do localadj^[i,j] := notthere;
         end;
       for i := 1 to max do 
            if (i in cluster[curpos]) and (i <> last) then localadj^[i,last] := 0;
       formbranch(localadj,tempcats,last);
       curpos := curpos + 1; 
     end;
   fillouttree(cats,MLinternal,0);
   dispose(localadj);  
end;

procedure convert(var where: wherearray; var evidencelist: observationlist; level: integer; var evidenceset: vertexset);
var i: integer;
begin
   removeallevidence;
   initev;
   evidenceset := [];
   for i := 1 to level - 1 do 
         addevidence(evidencelist,evidenceset,vernames[where[i].varnum],where[i].varnum,where[i].val);
end;

procedure insertlatentdata(vertex: vertextype; var category: probptr; level: integer; 
                           valuesofar: real; var data: vertexarray; var tempcats: categoryarray; 
                           var where: wherearray);
var curvar,j:vertextype; 
begin
   curvar := category^.varnum;
   where[level].varnum := curvar;
   if curvar <= numob then
      begin
         category^.count[data[curvar]] := category^.count[data[curvar]] + 1;
         where[level].val := round(cats[curvar].value[data[curvar]]);
      end
   else 
      begin
         if category^.multiplier[1] < 0 then
             begin
                convert(where,evidencelist,level,evidenceset);
                updatebayes;
                for j := 1 to maxval[curvar] do 
                   category^.multiplier[j] := exactfreqcard[curvar,j];
             end;
          for j := 1 to maxval[curvar] do 
               category^.count[j] := category^.count[j] + category^.multiplier[j] * valuesofar;
      end;
   if curvar <> vertex then
    if (curvar > numob) and (category^.nextvalue[1] <> nil) then
     for j := 1 to maxval[category^.nextvalue[1]^.varnum] do
      begin
        where[level].val := j;
        data[curvar] := j;
        insertlatentdata(vertex,category^.nextvalue[j],level+1,
                         valuesofar * category^.multiplier[j],data,tempcats,where);
      end
    else insertlatentdata(vertex,category^.nextvalue[data[curvar]],level+1,valuesofar,data,tempcats,where);
end;      


procedure filloutlatenttree(var category: categoryarray);
var i: integer; j: vertextype; datapoint:vertexarray; allknown: boolean;
begin
   for i := 1 to tround(sampsize) do 
     begin
        allknown := true;
	for j:= 1 to numob do
           begin
              datapoint[j]:=getbigdata(i,j);
              if datapoint[j] = 0 then allknown := false;
           end;
        for j := numob + 1 to max do
           if (category[j].condprob <> nil) and allknown then
              insertlatentdata(j,category[j].condprob,1,1,datapoint,category,where);
     end;
end;

procedure recalclatenttree(vertex: vertextype; var category: probptr; level: integer; 
                           prevcount: real; var data: vertexarray; var maxchange: real;
                           var where: wherearray);
var curvar,j:vertextype; diff: real; 
begin
   curvar := category^.varnum;
   where[level].varnum := curvar;
   if curvar > numob then
      begin
        convert(where,evidencelist,level,evidenceset);
        updatebayes;
        for j := 1 to maxval[curvar] do 
           begin
              diff := abs(category^.multiplier[j] - exactfreqcard[curvar,j]);
              category^.multiplier[j] := exactfreqcard[curvar,j];
              category^.count[j] := category^.multiplier[j] * prevcount;
              if diff > maxchange then maxchange := diff;
           end;
      end;
   if curvar <> vertex then
    for j := 1 to maxval[curvar] do
      begin
        data[curvar] := j;
        where[level].val := round(cats[curvar].value[j]);
        recalclatenttree(vertex,category^.nextvalue[j],level+1,category^.count[j],data,maxchange,where);
      end;
end;      

procedure recalctrees(var tempcats: categoryarray; var maxchange: real);
var i,j: vertextype; where: wherearray; data: vertexarray; 
begin
   for i := numob + 1 to max do 
      if tempcats[i].condprob <> nil then
          recalclatenttree(i,tempcats[i].condprob,1,1,data,maxchange,where);
end;

procedure insertthedata(dataval: real; vertex: vertextype; var category: probptr; 
                     var data: vertexarray; var cats: categoryarray);
var curvar:vertextype;
begin
   curvar := category^.varnum;
   if curvar = vertex then
      category^.count[data[vertex]] :=
         category^.count[data[vertex]] + dataval
   else insertthedata(dataval,vertex,category^.nextvalue[data[curvar]],data,cats);
end;      

procedure convertlatentbranch(vertex: vertextype; var category: probptr; level: integer; 
                              var cats: categoryarray; var data: vertexarray; 
                              var vertices: vertexset);
var curvar,i,j: vertextype;
begin
   curvar := category^.varnum;
   if curvar <> vertex then
     for j := 1 to maxval[category^.nextvalue[1]^.varnum] do
       begin
          vertices := vertices + [curvar];
          data[curvar] := j;
          convertlatentbranch(vertex,category^.nextvalue[j],level+1,cats,data,vertices);
       end
   else for j := 1 to max do
      if (j in vertices) then
          for i := 1 to maxval[vertex] do 
             begin
               data[curvar] := i;
               insertthedata(category^.count[i],j,cats[j].condprob,data,cats);
             end;
end;      

procedure renormalizelatent(curpos: probptr; v: vertextype);
var cut: real; i: vertextype; total: real;
begin
   if curpos <> nil then
    if curpos^.nextvalue[1] <> nil then
      begin
        i := 1;
        while curpos^.nextvalue[i] <> nil do 
          begin
            renormalizelatent(curpos^.nextvalue[i],i);
            i := i + 1;
          end;
      end
    else 
      begin
        total := 0;
        for i := 1 to maxval[v] do total := curpos^.count[i] + total;
        if total <> 0 then
          for i := 1 to maxval[v]-1 do 
             curpos^.cutoff[i] :=  curpos^.count[i]/total + curpos^.cutoff[i-1]
        else for i := 1 to maxval[v]-1 do
             curpos^.cutoff[i] := (1/maxval[v]) + curpos^.cutoff[i-1];
        curpos^.cutoff[maxval[v]] := 1.0;
      end;
end;

procedure convertlatenttree(var tempcats,cats: categoryarray);
var data: vertexarray; i: vertextype; vertices: vertexset;
begin
  for i := 1 to max do 
     if (i in haslatentpar) then initcount(cats[i].condprob,i);
  for i := numob + 1 to max do 
    if tempcats[i].condprob <> nil then
        begin
          vertices := [i];
          convertlatentbranch(i,tempcats[i].condprob,1,cats,data,vertices);
        end;
  for i := 1 to max do
    if (i in haslatentpar) then
       renormalizelatent(cats[i].condprob,i);
end;

procedure gettheprob(var sample: vertarray; var evidenceset: vertexset; var prob,count: real; ptr: probptr);
var curvar,i: vertextype;
begin
   i := 1;
   if ptr^.nextvalue[1] <> nil then
     while (ptr^.nextvalue[i] <> nil) and (i <= maxcategories) do
        begin
           curvar := ptr^.varnum;
           evidenceset := evidenceset + [curvar];
           sample[curvar] := cats[curvar].value[i];
           gettheprob(sample,evidenceset,prob,count,ptr^.nextvalue[i]);
           i := i + 1;
        end
   else 
     begin
         curvar := ptr^.varnum;
         for i := 1 to maxval[curvar] do
            begin
              evidenceset := evidenceset + [curvar];
              sample[curvar] := cats[curvar].value[i];
              count := count + ptr^.count[i];
              prob := prob + getlikelihoodscore(sample,evidenceset);
            end;
     end;
end;

procedure loglikelattree(ptr: probptr; var loglike: real; var sample: vertarray; 
                         var evidenceset: vertexset);
var curvar,i: vertextype; aprob,count: real;
begin
  curvar := ptr^.varnum;
  i := 1;
  if ptr^.nextvalue[1] <> nil then
      while (ptr^.nextvalue[i] <> nil) and (i <= maxcategories) do
         begin
           sample[curvar] := cats[curvar].value[i];
           evidenceset := evidenceset + [curvar];
           if ptr^.nextvalue[i]^.varnum <= numob then
               loglikelattree(ptr^.nextvalue[i],loglike,sample,evidenceset)
           else
              begin
                 aprob := 0;
                 count := 0;
                 gettheprob(sample,evidenceset,aprob,count,ptr^.nextvalue[i]);
                 loglike := loglike + ln(aprob) * count;
              end;
           i := i + 1;
         end;
end;

              
procedure loglikelihood(curpos: probptr; v: vertextype; var loglike: real);
var i: vertextype; aprob: real; 
begin
   if curpos <> nil then
    if curpos^.nextvalue[1] <> nil then
      begin
        i := 1;
        while (curpos^.nextvalue[i] <> nil) and (i <= maxcategories) do 
          begin
            loglikelihood(curpos^.nextvalue[i],i,loglike);
            i := i + 1;
          end;
      end
    else for i := 1 to maxval[v] do 
      begin
         aprob := curpos^.cutoff[i] - curpos^.cutoff[i-1];
         if aprob <> 0 then loglike := loglike + curpos^.count[i] * ln(aprob)
         else writeln('found 0 at ',v:3);
      end;
end;

procedure calcloglike(var cats: categoryarray; eqs: eqorder; var loglike: real);
var i: vertextype; notdone: vertexset; curpos: eqorder;
    sample: vertarray;
begin
   loglike := 0;
   notdone := [1..numob];
   for i := numob + 1 to max do
      if tempcats[i].condprob <> nil then 
         begin
            evidenceset := [];
            loglikelattree(tempcats[i].condprob,loglike,sample,evidenceset);
            notdone := notdone - evidenceset;
         end;
   curpos := eqs^.next;
   while curpos <> nil do
      begin
        if curpos^.varbl in notdone then
           loglikelihood(cats[curpos^.varbl].condprob,curpos^.varbl,loglike);
        curpos := curpos^.next;
      end;
   writeln(output,'loglike = ',loglike:8:4);
end;


procedure em(* var tempcats,cats: categoryarray; var score: real *);
var maxchange: real; eqs: eqorder; stopunit: real; i: vertextype;
begin
  stopunit := 0.001;
  new(eqs);
  formeqsorder(eqs,adjacency);
  initlatents(adjacency);
  getlatentclust(adjacency); 
  filloutlatenttree(tempcats);
  convertlatenttree(tempcats,cats); 
  if mydebug then writediscretemodel(cats,eqs,adjacency,output,false);
  repeat
     maxchange := 0.0;
     recalctrees(tempcats,maxchange);
     convertlatenttree(tempcats,cats);
     if mydebug then writediscretemodel(cats,eqs,adjacency,output,false);
  until maxchange < stopunit;
  calcloglike(cats,eqs,score);
  (* writediscretemodel(cats,eqs,adjacency,ofilvar,false);   *)
  disposeeqsorder(eqs);
  disposebayesnet(cats);
  disposebayesnet(tempcats);
  for i := 1 to max do
     begin
        if cats[i].condprob <> nil then dispose(cats[i].condprob);
        if tempcats[i].condprob <> nil then dispose(tempcats[i].condprob);
     end;
end;

procedure writetempset(var t: tempset);
var num,i: vertextype;
begin
  write(output,'Number: ');
  readln(input,num);
  for i := 1 to max do
     begin
       if i in t[num] then 
         begin
          writename(i,output);
          writeln(output);
         end;
     end;
end;

procedure collectprior;
var ofilename: inbuf; changed: boolean; confound,direct: real;
begin
   getknowledge;
   getfile(ofilvar,'Output file: |                          ',outputf,ofilename,changed);
   confound := rho^[2,1] * rho^[3,1];
   direct := rho^[2,3] - confound;
   writeln(ofilvar,confound:8:3,' ',direct:8:3);
end;

