#include "syst.h" 
#include "tt1.h" 
#include "tt2.h" 
#include "tt3.h"
#include "tt4.h" 
#include "tt5.h" 
#include "io.h"
#include "part.h"
#include "recalc.h"
#include "numeric.h"
#include "calctet.h"
#include "matrices.h"
#include "powell.h"
#include "sys.h"
#include "testbuildm.h"
#include "mon.h"
#include "linbayes.h"

const
   ftol=1.0e-6;
type
   glnarray = array [1..maxvertex] of real;
   glndim = glnarray;
   glnpbynp = array [1..maxvertex,1..maxvertex] of real;
   gl3array = glnarray;
   pindexrec = record
                 cause,effect,atype: vertextype;
               end;
   pindexarr = array[1..10000] of pindexrec;  

var
   ncom : integer;
   pcom,xicom : glnarray; 
   fret : real;
   iter,npp : integer;
   p : glnarray;
   xi : glnpbynp;
   pindex: pindexarr;


function neglog(var cov,tcov: covarray; var aux: auxiliary): real;
label 999;
var i,j: vertextype; tempnum1,tempnum2: real; 
begin
with aux do
  begin
  for i:= 1 to numob do
    for j:=1 to numob do
     tempcov^[i,j]:= tcov^[i,j];
  inverse(tempcov,theinverse,numob,aux.ok);  (* puts inverse of imp.covs in ginverse *)
  if (not aux.ok) then goto 999;
  sqmatmult(cov,theinverse,tempcov,numob);  (* stores answer in tcov *)
  tempnum1 := determ(tcov,numob,aux.ok);
  if (not aux.ok) then goto 999;
  tempnum2 := trace(tempcov,numob);
  if tempnum1 < 0 then 
       begin
(*         neglog := -1; *)
         aux.ok := false;
       end
  else neglog := (sampsize/2) * (ln(tempnum1) + trace(tempcov,numob)); 
(*  else neglog:= (sampsize/2)*(-1 * ln(tempnum1) + tempnum2);  *)
end;
999: end;

procedure getpathcovs(var covpaths,parm: covarray);
var j,k,i: integer;
begin (* calcpath *)
    for i := 1 to max do
       for j := 1 to max do 
              if (i <> j) and (parm^[i,j] <> parm^[j,i]) then covpaths^[i,j] := parm^[i,j]
              else covpaths^[i,j] := 0.0;
    for k := 1 to numob do
       for i := 1 to numob do
           for j := 1 to numob do
                 covpaths^[i,j] := covpaths^[i,j] + covpaths^[i,k] * covpaths^[k,j];
end; (* Calcdpath *)

procedure across(x,y: vertextype; var parm,tcov: covarray; calcvariances: boolean; var aux: auxiliary);
(* variances of error terms go in tcov^[i,i] *)
begin
  with theorder do
   begin
        for i := 1 to tierpos[x] - 1 do 
           for j := 1 to tierpos[y] - 1 do 
             if (x <> y) or (i <= j) then
               begin
                 if (calcvariances) or (tier[x,i] <> tier[y,j]) then 
                    begin
                        for k := 1 to numparents[tier[y,j]] do 
                          begin
(*                            writeln('part2 ',x,' ',y,' ',i,' ',j,' ',k,' ',tier[x,i],' ',tier[y,j],' ',parents[tier[y,j],k],' ',tcov^[tier[x,i],tier[y,j]],' ',
                                   parm^[parents[tier[y,j],k],tier[y,j]],' ',tcov^[tier[x,i],parents[tier[y,j],k]],' ',
                                  tcov^[tier[x,i],parents[tier[y,j],k]] * parm^[parents[tier[y,j],k],tier[y,j]]);  *)
                            tcov^[tier[x,i],tier[y,j]] := 
                               tcov^[tier[x,i],tier[y,j]] +
                                  tcov^[tier[x,i],parents[tier[y,j],k]] *
                                  parm^[parents[tier[y,j],k],tier[y,j]];
                            tcov^[tier[y,j],tier[x,i]] := tcov^[tier[x,i],tier[y,j]];
                          end;
                        for k := 1 to numdouble[tier[y,j]] do 
                           if (tier[x,i] <> tier[y,j]) and (tier[x,i] <> double[tier[y,j],k]) then
                               begin
(*                            writeln('part3 ',x,' ',y,' ',i,' ',j,' ',k,' ',tier[x,i],' ',tier[y,j],' ',double[tier[y,j],k],' ',tcov^[tier[x,i],tier[y,j]],' ',
                                   parm^[double[tier[y,j],k],tier[y,j]],' ',aux.tempcov^[double[tier[y,j],k],tier[x,i]],' ',
                                  aux.tempcov^[double[tier[y,j],k],tier[x,i]]  * parm^[double[tier[y,j],k],tier[y,j]]);   *)
                            tcov^[tier[x,i],tier[y,j]] := 
                               tcov^[tier[x,i],tier[y,j]] +
                                  aux.tempcov^[double[tier[y,j],k],tier[x,i]] *
                                  parm^[double[tier[y,j],k],tier[y,j]];
                            tcov^[tier[y,j],tier[x,i]] := tcov^[tier[x,i],tier[y,j]];
                          end;
                     end;
              end;
    end
end;

procedure formimp(var parm,tcov: covarray; calcvariances: boolean; var aux: auxiliary);
var i,j: vertextype;
begin
   getpathcovs(aux.tempcov,parm);
   for i := 1 to max do
     for j := 1 to max do
        if  (parm^[i,j] = parm^[j,i]) then tcov^[i,j] := parm^[i,j]
        else tcov^[i,j] := 0;
   for i := 1 to theorder.maxtiers do
     for j := 1 to i do
        across(j,i,parm,tcov,calcvariances,aux);
end;

function doubleheaded(i,j: vertextype): boolean;
begin
   doubleheaded := (parm^[i,j] <> 0) and (parm^[i,j] = parm^[j,i]) and (i <> j);
end;

function findparam(i: integer): integer;
var curpos: integer;
begin
   curpos := 1;
   while (pindex[curpos].atype <> 1) or (pindex[curpos].cause <> i) do 
      curpos := curpos + 1;
   findparam := curpos;
end;

function fnc(x: gl3array; var cov,parm,tcov: covarray; var aux: auxiliary): real;
(* make sure parm is initialized to zero *)
label 999;
var i,j: vertextype; tempnum: real; foundone: boolean; k: integer;
begin
   for i := 1 to ncom do
      if pindex[i].atype = 1 then 
         begin
            tempnum := expon(2.0,x[i]);
            parm^[pindex[i].cause,pindex[i].effect] := tempnum;
         end
      else if pindex[i].atype = 2 then 
          parm^[pindex[i].cause,pindex[i].effect] := x[i]
      else if pindex[i].atype = 3 then
        begin
          parm^[pindex[i].cause,pindex[i].effect] := x[i];
          parm^[pindex[i].effect,pindex[i].cause] := x[i];
        end;
   formimp(parm,tcov,true,aux);
   tempnum := neglog(cov,tcov,aux);
   if (not aux.ok) then
      begin
          tempnum := 100000000;
          goto 999;   
      end;
(*   while (tempnum < 0) do
     begin *)
       i := 1; 
       while (i <= max) and (tempnum < 0) do
         begin
            firsttime := false;
            j := 1;
            foundone := false;
            while (j <= max) and (not foundone) do
               begin
                  if doubleheaded(i,j) then 
                    begin
                      k := findparam(i);
                      x[k] := 2 * x[k];
                      parm^[i,i] := 2 * parm^[i,i];
                      foundone := true;
                    end
                  else j := j + 1;
                end;
            if foundone then
               begin
                  formimp(parm,tcov,true,aux);
                  tempnum := neglog(cov,tcov,aux);
               end;
            i := i + 1;
(*        end; *)
     if tempnum < 0 then 
        begin
           aux.ok := false;
           tempnum := 100000000;
        end;
    end; 
999:  fnc := tempnum;
end;

function f1dim(x: real; var cov,parm,tcov: covarray; var aux: auxiliary): real;
var
   j: integer;
   xt: glndim;
begin
   for j := 1 to ncom do begin
      xt[j] := pcom[j]+x*xicom[j]
   end;
   f1dim := fnc(xt,cov,parm,tcov,aux)
end;

function func(x: real; var cov,parm,tcov: covarray; var aux: auxiliary): real;
begin
   func := f1dim(x,cov,parm,tcov,aux);
end;

procedure mnbrak(var ax,bx,cx,fa,fb,fc: real; var cov,parm,tcov: covarray; var aux: auxiliary);
label 1,999;
const
   gold=1.618034;
   glimit=100.0;
   tiny=1.0e-20;
var
   ulim,u,r,q,fu,dum: real;
function max(a,b: real): real;
   begin
      if (a > b) then max := a else max := b
   end;
function sign(a,b: real): real;
   begin
      if (b > 0.0) then sign := abs(a) else sign := -abs(a)
   end;
begin
   fa := func(ax,cov,parm,tcov,aux);
(*   if (not aux.ok) then goto 999;   *)
   fb := func(bx,cov,parm,tcov,aux);
(*   if (not aux.ok) then goto 999;  *)
   if (fb > fa) then begin
      dum := ax;
      ax := bx;
      bx := dum;
      dum := fb;
      fb := fa;
      fa := dum
   end;
   cx := bx+gold*(bx-ax);
   fc := func(cx,cov,parm,tcov,aux);
(*   if (not aux.ok) then goto 999; *)
1:   if (fb >= fc) then begin
      r := (bx-ax)*(fb-fc);
      q := (bx-cx)*(fb-fa);
      u := bx-((bx-cx)*q-(bx-ax)*r)/
         (2.0*sign(max(abs(q-r),tiny),q-r));
      ulim := bx+glimit*(cx-bx);
      if ((bx-u)*(u-cx) > 0.0) then begin
         fu := func(u,cov,parm,tcov,aux);
(*         if (not aux.ok) then goto 999; *)
         if (fu < fc) then begin
            ax := bx;
            fa := fb;
            bx := u;
            fb := fu;
            goto 1 end
         else if (fu > fb) then begin
            cx := u;
            fc := fu;
            goto 1
         end;
         u := cx+gold*(cx-bx);
         fu := func(u,cov,parm,tcov,aux);
         if (not aux.ok) then goto 999;
      end else if  ((cx-u)*(u-ulim) > 0.0) then begin
         fu := func(u,cov,parm,tcov,aux);
(*         if (not aux.ok) then goto 999; *)
         if (fu < fc) then begin
            bx := cx;
            cx := u;
            u := cx+gold*(cx-bx);
            fb := fc;
            fc := fu;
            fu := func(u,cov,parm,tcov,aux);
(*            if (not aux.ok) then goto 999; *)
         end
      end else if  ((u-ulim)*(ulim-cx) >= 0.0) then begin
         u := ulim;
         fu := func(u,cov,parm,tcov,aux);
(*         if (not aux.ok) then goto 999; *)
      end else begin
         u := cx+gold*(cx-bx);
         fu := func(u,cov,parm,tcov,aux);
(*         if (not aux.ok) then goto 999; *)
      end;
      ax := bx;
      bx := cx;
      cx := u;
      fa := fb;
      fb := fc;
      fc := fu;
      goto 1
   end;
999: end;

function mybrent(ax,bx,cx,tol: real; var xmin: real; var cov,parm,tcov: covarray; var aux: auxiliary): real;
label 1,2,3,999;
const
   itmax=100;
   cgold=0.3819660;
   zeps=1.0e-10;
var
   a,b,d,e,etemp: real;
   fu,fv,fw,fx: real;
   iter: integer;
   p,q,r,tol1,tol2: real;
   u,v,w,x,xm: real;
function sign(a,b: real): real;
   begin
      if (b > 0.0) then sign := abs(a) else sign := -abs(a)
   end;
begin
   if ax < cx then a := ax else a := cx;
   if ax > cx then b := ax else b := cx;
   v := bx;
   w := v;
   x := v;
   e := 0.0;
   fx := func(x,cov,parm,tcov,aux);
(*   if (not aux.ok) then goto 999; *)
   fv := fx;
   fw := fx;
   for iter := 1 to itmax do begin
      xm := 0.5*(a+b);
      tol1 := tol*abs(x)+zeps;
      tol2 := 2.0*tol1;
      if (abs(x-xm) <= (tol2-0.5*(b-a))) then goto 3;
      if (abs(e) > tol1) then begin
         r := (x-w)*(fx-fv);
         q := (x-v)*(fx-fw);
         p := (x-v)*q-(x-w)*r;
         q := 2.0*(q-r);
         if (q > 0.0) then  p := -p;
         q := abs(q);
         etemp := e;
         e := d;
         if((abs(p) >= abs(0.5*q*etemp)) or (p <= q*(a-x))
            or (p >= q*(b-x))) then goto 1;
         d := p/q;
         u := x+d;
         if (((u-a) < tol2) or ((b-u) < tol2)) then d := sign(tol1,xm-x);
         goto 2
      end;
1:      if (x >= xm)  then e := a-x else e := b-x;
      d := cgold*e;
2:      if (abs(d) >= tol1)  then u := x+d else u := x+sign(tol1,d);
      fu := func(u,cov,parm,tcov,aux);
(*      if (not aux.ok) then goto 999; *)
      if (fu <= fx)  then begin
         if (u >= x)  then a := x else b := x;
         v := w;
         fv := fw;
         w := x;
         fw := fx;
         x := u;
         fx := fu
      end else begin
         if (u < x)  then a := u else b := u;
         if ((fu <= fw) or (w = x))  then begin
            v := w;
            fv := fw;
            w := u;
            fw := fu
         end else if ((fu <= fv) or (v = x) or (v = 2)) then begin
            v := u;
            fv := fu
         end
      end
   end;
   writeln('pause in routine brent - too many iterations');
   mybrent := 100000000;
   xmin := 10000000;
   goto 999;
3:   xmin := x;
   mybrent := fx;
999: end;

procedure linmin(var p,xi: glnarray; n: integer; var fret: real; var cov,parm,tcov: covarray;
                 var aux: auxiliary);
label 999; 
const
   tol=1.0e-4;
var
   j: integer;
   xx,xmin,fx,fb,fa,bx,ax: real;
begin
   ncom := n;
   for j := 1 to n do begin
      pcom[j] := p[j];
      xicom[j] := xi[j]
   end;
   ax := 0.0;
   xx := 1.0;
   bx := 2.0;
   mnbrak(ax,xx,bx,fa,fx,fb,cov,parm,tcov,aux);
(*   if (not aux.ok) then goto 999; *)
   fret := mybrent(ax,xx,bx,tol,xmin,cov,parm,tcov,aux);
(*   if (not aux.ok) then goto 999; *)
   for j := 1 to n do begin
      xi[j] := xmin*xi[j];
      p[j] := p[j]+xi[j]
   end;
999: end;

procedure powell(var p: glnarray; var xi: glnpbynp; n,npp: integer;
       ftol: real; var iter: integer; var fret: real; var cov,parm,tcov: covarray; var aux: auxiliary);
label 1,999;
const
   itmax=200;
var
   j,ibig,i: integer;
   t,fptt,fp,del: real;
   pt,ptt,xit: glnarray;
begin
   ncom := n;
   fret := fnc(p,cov,parm,tcov,aux);
(*   if (not aux.ok) then goto 999; *)
   for j := 1 to n do begin
      pt[j] := p[j]
   end;
   iter := 0;
1: iter := iter+1;
   fp := fret;
   ibig := 0;
   del := 0.0;
   for i := 1 to n do begin
      for j := 1 to n do begin
         xit[j] := xi[j,i]
      end;
      fptt := fret;
      linmin(p,xit,n,fret,cov,parm,tcov,aux);
      if (abs(fptt-fret) > del) then begin
         del := abs(fptt-fret);
         ibig := i
      end
   end;
   if (2.0*abs(fp-fret) <= ftol*(abs(fp)+abs(fret))) then goto 999;
   if (iter = itmax) then begin
      writeln('pause in routine powell');
      writeln('too many interations'); readln
   end;
   for j := 1 to n do begin
      ptt[j] := 2.0*p[j]-pt[j];
      xit[j] := p[j]-pt[j];
      pt[j] := p[j]
   end;
   fptt := fnc(ptt,cov,parm,tcov,aux);
(*   if (not aux.ok) then goto 999; *)
   if (fptt >= fp) then goto 1; 
   t := 2.0*(fp-2.0*fret+fptt)*sqr(fp-fret-del)-del*sqr(fp-fptt);
   if (t >= 0.0) then goto 1; 
   linmin(p,xit,n,fret,cov,parm,tcov,aux);
   for j := 1 to n do begin
      xi[j,ibig] := xit[j]
   end;
   goto 1;
999:   end;

procedure putintier(x,tiernum: vertextype; var thenext: vertextype);
var i: vertextype;
begin
  with theorder do
   begin
    tier[tiernum,tierpos[tiernum]] := x;
    tierpos[tiernum] := tierpos[tiernum] + 1;
    for i := 1 to numchildren[x] do
       begin
          numparents1[children[x,i]] := numparents1[children[x,i]] - 1;
          if numparents1[children[x,i]] = 0 then
              begin
                 nextinline[thenext].avar := children[x,i];
                 nextinline[thenext].tier := tiernum + 1;
                 if (tiernum + 1) > maxtiers then maxtiers := tiernum + 1;
                 thenext := thenext + 1;
              end;
       end;
     end;
end;

procedure myeqsorder(var mag: graphrec);
var i,j,thenext: vertextype;
begin
 theorder.maxtiers := 1;
 thenext := 1;
 with theorder do
  begin
   for i := 1 to max do
      begin
         tierpos[i] := 1;
         numparents[i] := 0;
         numparents1[i] := 0;
         numchildren[i] := 0;
         numdouble[i] := 0;
      end;
   for i := 1 to max do
     for j := 1 to max do
       begin
          if (in2 in mag.augraph^[i,j]) and (out2 in mag.augraph^[j,i]) then
             begin
                numparents[j] := numparents[j] + 1;
                numparents1[j] := numparents1[j] + 1;
                numchildren[i] := numchildren[i] + 1;
                children[i,numchildren[i]] := j;
                parents[j,numparents[j]] := i;
             end
          else if (in2 in mag.augraph^[i,j]) and (in2 in mag.augraph^[j,i]) and (i < j) then
             begin
                numdouble[i] := numdouble[i] + 1;
                numdouble[j] := numdouble[j] + 1;
                double[i,numdouble[i]] := j;
                double[j,numdouble[j]] := i;
             end;
       end;
    for i := 1 to max do if (numparents[i] = 0) then putintier(i,1,thenext);
    i := 1;
    while (i < thenext) do
       begin
         putintier(nextinline[i].avar,nextinline[i].tier,thenext);
         i := i + 1;
       end;
  end;
end;

procedure initscore(var mag: graphrec; var parm: covarray);
var i,j: vertextype;
begin
   npp := 0;
   myeqsorder(mag);
   with mag do 
     begin
       for i := 1 to max do 
         for j := 1 to max do
            begin
              if (i = j) (* and theorder.numchildren[i] <> 0) *) then
                begin
                  npp := npp + 1;
                  p[npp] := 1;
                  pindex[npp].cause := i;
                  pindex[npp].effect := j;
                  pindex[npp].atype := 1;
                  parm^[i,j] := 1;
                end
              else if (in2 in augraph^[i,j]) and (out2 in augraph^[j,i]) then
                begin
                  npp := npp + 1;
                  p[npp] := 0.1;
                  pindex[npp].cause := i;
                  pindex[npp].effect := j;
                  pindex[npp].atype := 2;
                  parm^[i,j] := 0.1;
                end
              else if (in2 in augraph^[i,j]) and (in2 in augraph^[j,i]) and (i < j) then
                begin
                  npp := npp + 1;
                  p[npp] := 0.1;
                  pindex[npp].cause := i;
                  pindex[npp].effect := j;
                  pindex[npp].atype := 3;
                  parm^[i,j] := 0.1;
                  parm^[j,i] := 0.1;
                end
              else if (not ((in2 in augraph^[i,j]) and (in2 in augraph^[j,i]))) then parm^[i,j] := 0;
            end;
       for i := 1 to npp do 
         for j := 1 to npp do
           if i = j then xi[i,j] := 1.0 else xi[i,j] := 0;
      end;
end;       
    
procedure scorecluster(var mag: graphrec; var score: real; var ok: boolean;
                       var cov,parm,tcov: covarray; var likeli: real; var aux: auxiliary);
label 999; 
var ndim: integer; 
begin
   initscore(mag,parm);
   ndim := npp;
   powell(p,xi,ndim,npp,ftol,iter,fret,cov,parm,tcov,aux);
   if (not aux.ok) then goto 999;
   likeli := fret;
   score := -1 * (fret - 1/2 * ndim * ln(sampsize));
999: end;

procedure printxi;
var i,j: vertextype;
begin
   for i := 1 to npp do
      for j := 1 to npp do
        write('i = ',i,' j = ',j,' xi = ',xi[i,j]);
end;

function insamecluster(v1,v2: vertextype; var cluster: clustertype): boolean;
begin
   i := 1;
   while (i <= max) and (cluster[i].verset <> []) and (not (v1 in cluster[i].verset)) and
         (not (v2 in cluster[i].verset)) do i := i + 1;
   insamecluster := (v1 in cluster[i].verset) and (v2 in cluster[i].verset);
end;

procedure readmagtoadj(var mag: graphrec; var tempadj: adjtype; var cluster: clustertype;
                       var aux: auxiliary);
var i,j: vertextype;
begin
   for i := 1 to max do 
     for j := 1 to max do
        begin
           if (not insamecluster(i,j,cluster)) then
              if (in2 in mag.augraph^[i,j]) and (out2 in mag.augraph^[j,i]) then
                 begin
                    tempadj^[i,j] := 0;
                    tempadj^[j,i] := notthere;
                 end
              else if (in2 in mag.augraph^[j,i]) and (out2 in mag.augraph^[i,j]) then
                 begin
                   tempadj^[j,i] := 0;
                   tempadj^[i,j] := notthere;
                 end
              else 
                 begin
                   tempadj^[j,i] := notthere;
                   tempadj^[i,j] := notthere;
                 end           
          else
             begin
               tempadj^[j,i] := notthere;
               tempadj^[i,j] := notthere;
             end;
        end;
end;


procedure chaindouble(var mag: graphrec; var doublehead: adjtype; var cluster: clustertype; 
                      var aux: auxiliary);
var i,j,k,l,m,current: vertextype; visited: vertexset; addedone: boolean;
    expandlist: array[1..maxvertex] of vertextype;
begin
with aux do 
 begin
  incluster := [];
  visited := [];
  for j := 1 to max do 
     begin
       cluster[j].verset := [];
       cluster[j].ancestset := [];
       cluster[j].descend := [];
       cluster[j].parentset := [];
     end;
  j := 1;
  for i := 1 to max do
   begin
    addedone := false;
    if (doublehead^[i,1] <> 0) and (not (i in visited)) then
       begin
         addedone := true;
         current := i;
         cluster[j].verset := cluster[j].verset + [i];
         incluster := incluster + [i];
         cluster[j].parentset := parentset[i];
         cluster[j].ancestset := mag.ancestset[i];
         cluster[j].descend := mag.descend[i];
         m := 1;
         l := 1;
         while current <> 0 do
            begin        
               visited := visited + [current];
               k := 1;
               while  (doublehead^[current,k] <> 0) do
                begin
                 if (not (doublehead^[current,k] in cluster[j].verset)) then
                   begin
                    cluster[j].verset := cluster[j].verset + [doublehead^[current,k]];
                    cluster[j].ancestset := cluster[j].ancestset + mag.ancestset[doublehead^[current,k]];
                    cluster[j].descend := cluster[j].descend + mag.descend[doublehead^[current,k]];
                    cluster[j].parentset := cluster[j].parentset + parentset[doublehead^[current,k]];
                    incluster := incluster + [doublehead^[current,k]];
                    expandlist[l] := doublehead^[current,k];
                    l := l + 1;
                   end;
                  k := k + 1;
                 end;
                if m < l then current := expandlist[m] else current := 0;
                expandlist[m] := 0;
                m := m + 1;
            end;
      end;
    visited := visited + [i];
    if addedone then j := j + 1;                
  end;
 end;
end;

function findcluster(v: vertextype; cluster: clustertype): vertextype;
var k: vertextype; 
begin
   k := 1;
   while (k <= max) and (not (v in cluster[k].verset)) do 
         k := k + 1;
   findcluster := k;
end;

procedure closedescend(var cluster: clustertype); var i,j: vertextype;
begin
   i := 2;
   while (i <= max) and (cluster[i].verset <> []) do
      begin
        j := 1;
        while (j < i) do
           begin
              if (cluster[j].verset * cluster[i].descend) <> [] then
                cluster[i].descend := cluster[i].descend + cluster[j].descend;
              if (cluster[j].verset * cluster[i].ancestset) <> [] then
                cluster[i].ancestset := cluster[i].ancestset + cluster[j].ancestset;
              if (cluster[i].verset * cluster[j].descend) <> [] then
                cluster[j].descend := cluster[j].descend + cluster[i].descend;
              if (cluster[i].verset * cluster[j].ancestset) <> [] then
                cluster[j].ancestset := cluster[j].ancestset + cluster[i].ancestset;
              j := j + 1;
           end;
        i := i + 1;
     end; end;

procedure mergeclusters(var mag: graphrec; var cluster: clustertype;
                        var aux: auxiliary); 
var foundone: boolean; i,j,k,index: vertextype; aset: vertexset; 
begin
 with aux do
 begin
  repeat
   foundone := false;
   i := 1;
   while (i <= max) and (cluster[i].verset <> []) do
      begin
         aset := (cluster[i].ancestset * cluster[i].descend) - cluster[i].verset;
         if aset <> [] then
            begin
               foundone := true;
               for j := 1 to max do
                  if (j in aset) then
                     begin
                        if (not (j in incluster)) then
                           begin
                              cluster[i].verset := cluster[i].verset + [j];
                              cluster[i].parentset := cluster[i].parentset + parentset[j];
                              cluster[i].ancestset := cluster[i].ancestset + mag.ancestset[j];
                              cluster[i].descend := cluster[i].descend + mag.descend[j];
                              incluster := incluster + [i];
                           end
                        else
                           begin
                              index := findcluster(j,cluster);
                              if (index <> i) then
                                 begin
                                    cluster[i].verset := cluster[i].verset + cluster[index].verset;
                                    cluster[i].descend := cluster[i].descend + cluster[index].descend;
                                    cluster[i].ancestset := cluster[i].ancestset + cluster[index].ancestset;
                                    cluster[i].parentset := cluster[i].parentset + cluster[index].parentset;
                                    k := index;
                                    while (cluster[k].verset <> []) and (k <= max) do
                                       begin
                                         cluster[k] := cluster[k+1];
                                         k := k + 1;
                                      end
                                end
                           end
                    end
            end;
          i := i + 1;
        end;
    if foundone then closedescend(cluster);
  until (not foundone);
 end;
end;
    
procedure getclusters(var mag: graphrec; var doublehead,parent: adjtype; var cluster: clustertype; 
                     var incluster: vertexset; var aux: auxiliary);
begin
   chaindouble(mag,doublehead,cluster,aux);
   closedescend(cluster);
   mergeclusters(mag,cluster,aux);
end;

procedure convertmag(var mag: graphrec; var parent,doublehead: adjtype; var numedges: integer;
                     var aux: auxiliary);
var i,j: vertextype; 
begin
 with aux do
  begin
   for i := 1 to max do 
       begin
         indexpar[i] := 1;
         indexdouble[i] := 1;
         parentset[i] := [];
         parent^[i,1] := 0;
         doublehead^[i,1] := 0;
       end;
   numedges := 0;
   for i := 1 to max do
       for j := 1 to max do
          if (in2 in mag.augraph^[i,j]) and (out2 in mag.augraph^[j,i]) then
               begin
                  parentset[j] := parentset[j] + [i];
                  parent^[j,indexpar[j]] := i;
                  indexpar[j] := indexpar[j] + 1;
                  numedges := numedges + 1;
               end
          else if (in2 in mag.augraph^[i,j]) and (in2 in mag.augraph^[j,i]) then
               begin
                  doublehead^[i,indexdouble[i]] := j;
                  indexdouble[i] := indexdouble[i] + 1;
                  if (i < j) then numedges := numedges + 1;
               end
  end;
end;

procedure readclusters(var mag: graphrec; cluster: vertexset; var tempmag: graphrec;
                       var newmax: vertextype; var map: parray; var cov,tcov: covarray);
var i,j: vertextype; mapinv: parray;
begin
   newmax := 1;
   for i := 1 to max do
     if (i in cluster) then 
       begin
          map[newmax] := i;
          mapinv[i] := newmax;
          newmax := newmax + 1;
       end;
   for i := 1 to max do
      if (i in cluster) then
        for j := 1 to max do
           if (j in cluster) then
             begin
                tempmag.augraph^[mapinv[i],mapinv[j]] := mag.augraph^[i,j];
                tempmag.augraph^[mapinv[j],mapinv[i]] := mag.augraph^[j,i];
                tcov^[mapinv[i],mapinv[j]] := cov^[i,j];
                tcov^[mapinv[j],mapinv[i]] := cov^[j,i];
             end;
   newmax := newmax - 1;
end;

procedure fillinparm(var parm,trueparm: covarray; var map: parray); 
begin
    for i := 1 to max do 
       for j := i to max do
          begin
             trueparm^[map[i],map[j]] := parm^[i,j];
             trueparm^[map[j],map[i]] := parm^[j,i];
          end;
end;

procedure getcov(var parent: adjtype; v: vertextype; var tcov,cov: covarray; var aux: auxiliary);
var i,j: vertextype;
begin
   i := 1;
   while (i < aux.indexpar[v]) do
     begin
       j := 1;
       while (j < aux.indexpar[v]) and (j <= i) do 
          begin
            tcov^[i,j] := cov^[parent^[v,i],parent^[v,j]];
            tcov^[j,i] := tcov^[i,j];
            j := j + 1;
          end;
       i := i + 1;
     end;
end;

procedure getcov1(aset: vertexset; var tcov,cov: covarray; var num: vertextype);
var i,j,index1,index2: vertextype;
begin
   index1 := 1;
   for i := 1 to max do
     if (i in aset) then 
       begin
          index2 := 1;
          for j := 1 to i do
             if j in aset then
                begin
                  tcov^[index1,index2] := cov^[i,j];
                  tcov^[index2,index1] := tcov^[index1,index2];
                  index2 := index2 + 1;
               end;
          index1 := index1 + 1;
       end;
   num := index1 - 1;
end;

procedure scoremag(* var mag: graphrec; var score: real; var aux: auxiliary *);
label 999;
var ok: boolean; temp1: real; dev: real;
    score1,likeli: real; num,newmax: vertextype; 
    map: parray; i,oldmax,oldnumob: vertextype; df: integer;
begin
 with aux do
  begin  
   convertmag(mag,parent,doublehead,numedges,aux);
   getclusters(mag,parent,doublehead,clusters,incluster,aux);
   readmagtoadj(mag,aux.tempadj,clusters,aux);
   linestimate(aux.tempadj,cov,localparm,ok);  
   i := 1;
   oldmax := max;
   oldnumob := numob;
   while (i <= max) and (clusters[i].verset <> []) do
      begin
         readclusters(mag,clusters[i].verset + clusters[i].parentset,apag,newmax,map,cov,localcov); 
         max := newmax;
         numob := newmax;
         scorecluster(apag,score1,ok,localcov,parm,estcov,likeli,aux); 
         clusters[i].score := -1 * likeli;
         fillinparm(parm,localparm,map); 
         max := oldmax;
         numob := oldnumob;  
         i := i + 1;
      end;
   for i := 1 to max do localparm^[i,i] := cov^[i,i];
   myeqsorder(mag);
   formimp(localparm,estcov,false,aux);
   score := 0;
   for i := 1 to max do
       begin
          if doublehead^[i,1] = 0 then
             begin
                lscorearr[i] := 0;
                if indexpar[i] <> 1 then 
                   begin
                      getcov(parent,i,localcov,cov,aux);
                      getcov(parent,i,localestcov,estcov,aux);
                      oldmax := max;
                      oldnumob := numob;
                      max := indexpar[i] - 1;
                      numob := indexpar[i] - 1;
                      lscorearr[i] := neglog(localcov,localestcov,aux);
                      if (not aux.ok) then goto 999;
                      max := oldmax;
                      numob := oldnumob;  
                   end;
                parent^[i,indexpar[i]] := i;
                indexpar[i] := indexpar[i] + 1;
                getcov(parent,i,localcov,cov,aux);
                getcov(parent,i,localestcov,estcov,aux);
                oldmax := max;
                oldnumob := numob;
                max := indexpar[i] - 1;
                numob := indexpar[i] - 1;
                lscorearr[i] := lscorearr[i] - neglog(localcov,localestcov,aux);
                if (not aux.ok) then goto 999;
                max := oldmax;
                numob := oldnumob;  
                indexpar[i] := indexpar[i] - 1;
                parent^[i,indexpar[i]] := 0;
                score := score + lscorearr[i];
             end;
         end;
   i := 1;
   while (i <= max) and (clusters[i].verset <> []) do
       begin
          getcov1(clusters[i].parentset - clusters[i].verset,localcov,cov,num);
          getcov1(clusters[i].parentset - clusters[i].verset,localestcov,estcov,num);
          oldmax := max;
          oldnumob := numob;
          max := num;
          numob := num;
          if num <> 0 then clusters[i].score := clusters[i].score + neglog(localcov,localestcov,aux);
          if (not aux.ok) then goto 999;
          score := score + clusters[i].score;
          max := oldmax;
          numob := oldnumob;  
          i := i + 1;
      end; 
   df := numedges + max;
(*   temp1 := neglog(cov,aux.estcov,aux);
   writeln(' temp1 = ',temp1,' score = ',score); 
   dev := neglog(cov,cov,aux);
   dev := dev - temp1;
   writeln('dev = ',dev);  *)
999:  if (aux.ok) then score := score - 1/2 * df * ln(sampsize)
       else score := -1000000000;
       max := oldmax;
       numob := oldnumob;  
 end;
end;

procedure printclusters(var clusters: clustertype);
var i,j: vertextype;
begin
  for j := 1 to max do
    if clusters[j].verset <> [] then 
      begin
       writeln('cluster number ',j);
       for i := 1 to max do
          begin
            if i in clusters[j].verset then 
               begin
                 writename(i,output);
                 writeln(output);
               end;
          end;
      end;
end;

procedure wcvo(var rho: covarray);
var i,j,numwritten: integer;
begin
    twrite(output,'  |');
    numwritten := 1;
    for i := 1 to numob do
       begin
          if (numwritten mod 8) = 0 then twriteln(output,'  |');
          pwritename(i,output,10);
               numwritten := numwritten + 1;
       end;
    numwritten := 1;
    for i := 1 to numob do 
        for j := 1 to numob do
            begin
               if (numwritten mod 8 = 0) or (j = 1) then 
                    begin
                       twriteeoln(output);
                       numwritten := 1;
                    end;
               tpwritereal(output,rho^[i,j],10,decpoints,false);
               numwritten := numwritten + 1;
            end;
    twriteeoln(output);
    twriteeoln(output);
end;

procedure wr(x: real);
begin
   writeln(output,x);
end;
