import java.io.*; 
import java.util.*; 
 
public class G2stat_test{ 
 
  /** 
   * how many samples in the data file 
   */ 
  public int sample_size; 
  /** 
   * The total number of variables (vertices) 
   */ 
  public int total_num_variable; 
  /** 
   * The data will be stored in a 2-d array, each row will store each sample 
   * For example, if we obersve 1,     2,     2,     3,     2 
   * in the nineth sample, the there will be 
   * data[8][0]=1, data[8][1]=2,data[8][2]=2, data[8][3]=3,data[8][4]=2 
   */ 
  protected int data[][]; 
  /** 
   * The significant level for the test, say 0.05 
   */ 
  public double signif; 
  /** 
   * The array to store all observed variables 
   */ 
  public Variable varArray[]; 
  /** 
   * The 1-d array to store the joint distribution table 
   */ 
  protected int cells[]; 
  /**  
   * The total number of cells which expected value is 0 
   */ 
  protected int num_zero_exp; 
  /** 
   * An array to store the test variable index 
   * for example if we want to test if y and v are independent given x 
   * then variable_index[0]=1, variable_index[1]=3, variable_index[2]=0 
   */ 
  protected int variable_index[]; 
  private int indices[]; 
 
  /** 
   * Set up the indices in order to transform multi-array to 1-d array 
   */ 
  protected void setupindices(){ 
     
      this.indices = new int[variable_index.length]; 
      this.indices[0]=1; 
      for(int i=1; i<variable_index.length ; i++) 
        this.indices[i]=this.indices[i-1]*varArray[variable_index[i-1]].size; 
  } 
 
/** 
 * Calculate hom many cells we need for the joint distribution table 
 */ 
 
  protected int totalCell(){ 
    int temp=1; 
    for(int i=0; i<variable_index.length; i++) 
      temp = temp*varArray[variable_index[i]].size; 
    return temp; 
  } 
 
/** 
 * given a sample, to dicide where it'll be in the joint distribution table 
 */ 
 
  protected int calclocation(int oneSample[]){ 
    int loc=0; 
    for(int i=0; i<variable_index.length; i++) 
      loc = loc + oneSample[variable_index[i]] * this.indices[i]; 
    return loc; 
  } 
 
/** 
 * Build up the joint distribution table for given data 
 */ 
 
  protected void fillincells(){ 
    int i,j,loc; 
    int[] temp = new int[total_num_variable]; 
    int totalcell=totalCell(); 
    this.cells = new int[totalcell]; 
    //initialize cells 
    for (i=0; i<totalcell; i++) this.cells[i]=0; 
 
System.out.println("sample_size "+ sample_size); 
System.out.println("total_num_variable "+ total_num_variable); 
 
    for(i=0; i<sample_size;i++){ 
      for (j=0; j<total_num_variable;j++) 
        temp[j]=data[i][j]; 
      loc = calclocation(temp); 
      this.cells[loc] = this.cells[loc]+1; 
    } 
  } 
 
/**  
 * Given a 1-D array index to calculate its multi-D indexes 
 */ 
 
  protected int[] reverseLoc(int loc){ 
    int[] series = new int[variable_index.length]; 
    for (int i=variable_index.length-1; i>0; i--){ 
      series[i]= loc/this.indices[i]; 
      loc = loc - (loc/this.indices[i])*this.indices[i]; 
    } 
    series[0] = loc; 
    return series; 
  } 
 
 
/** 
 * Given a multi-D indices to compute its 1-D location 
 */ 
 
  protected int calcLoc(int series[]){ 
    int loc = 0; 
    for (int i=0; i<variable_index.length; i++) 
      loc = loc + series[i]*this.indices[i]; 
    return loc; 
  } 
 
/** 
 * sum the first variable, ie X(+jk) 
 */ 
 
  protected int sumFirst(int series[]){ 
    int sum = 0; 
    int i; 
    int local_series[] = new int[series.length]; 
    for (i=0;i<series.length;i++) 
      local_series[i] = series[i]; 
 
    for (i=0; i<varArray[variable_index[0]].size; i++){ 
      local_series[0]=i; 
      sum = sum + this.cells[calcLoc(local_series)]; 
    } 
    return sum; 
  } 
 
/** 
 * sum the second variable, ie X(i+k) 
 */ 
 
  protected int sumSecond(int series[]){ 
    int sum = 0; 
    int i; 
    int local_series[] = new int[series.length]; 
    for (i=0;i<series.length;i++) 
      local_series[i] = series[i]; 
 
    for (i=0; i<varArray[variable_index[1]].size; i++){ 
      local_series[1]=i; 
      sum = sum + this.cells[calcLoc(local_series)]; 
    } 
    return sum; 
  } 
 
/** 
 * sum the both variables, ie X(++k) 
 */ 
 
  protected int sumBoth(int series[]){ 
    int sum = 0; 
    int i; 
    int local_series[] = new int[series.length]; 
    for (i=0;i<series.length;i++) 
      local_series[i] = series[i]; 
 
    for (i=0; i<varArray[variable_index[0]].size; i++){ 
      for (int j=0; j<varArray[variable_index[1]].size; j++){ 
        local_series[0]=i; 
        local_series[1]=j; 
        sum = sum + this.cells[calcLoc(local_series)]; 
      } 
    } 
    return sum; 
  } 
 
 
/** 
 * calclulate the G2 for given variable_index: first is "x_index",  
 * second is "y_index", and the residual are indices for conditional variables 
 */ 
 
  protected double calcg2(){ 
    double g2=0.0;  
    double observed, expected; 
    int total; 
    int[] series; 
    int j; 
 
    this.num_zero_exp =0; 
 
    for(int i=0; i<totalCell(); i++){ 
      series = reverseLoc(i); 
      total = sumBoth(series); 
      if (total == 0)  
        expected = 0.0; 
      else 
        expected =(double)sumFirst(series) * (double)sumSecond(series) / (double) total; 
 
      //System.out.println("sumFirst= "+sumFirst(series)); 
      //System.out.println("sumSecond "+sumSecond(series)); 
 
      observed =(double)this.cells[i]; 
      //System.out.println("observed = "+observed); 
      if (expected == 0.0)  
        this.num_zero_exp = this.num_zero_exp+1; 
      else if (observed != 0) 
        g2 = g2 + observed * Math.log(observed/expected); 
    } 
 
    return 2*g2; 
  } 
 
 
 
/** 
 * To check if the input has any error, such as the input String is not 
 * variable name and so on 
 */ 
 
  protected boolean isAnError(String x, String y, String[] condition){ 
    int i; 
 
    if(x.equals(y)) { 
      System.out.println("x and y can't be same, you can only calculate the independence of two different variables"); 
      return true; 
    } else if (condition != null){ 
      for(i=0; i< condition.length; i++){ 
        if (x.equals(condition[i]) || y.equals(condition[i])){ 
           System.out.println("The conditional variables must be different form x and y"); 
           return true; 
        } 
      } 
    }//end of else 
 
    Vector varSet = new Vector(total_num_variable); 
    for (i=0; i<total_num_variable; i++) 
      varSet.addElement(varArray[i].name); 
    if(!varSet.contains(x)){ 
      System.out.println("Wrong variable name, the first input "+x+" is not a variable name"); 
      return true; 
    } 
 
    if(!varSet.contains(y)){ 
      System.out.println("Wrong variable name, the second input"+y+" is not a variable name"); 
      return true; 
    } 
 
    if(condition != null){ 
System.out.println("Length of condition set T: "+condition.length);
      for(i=0; i< condition.length; i++){ 
        if(!varSet.contains(condition[i])){ 
         System.out.println("Wrong variable name, the number "+(i+1)+" String in condition set, "+condition[i]+", is not a variable name");  
         return true; 
        } 
      }//end for 
    } 
 
    return false; 
  } 
 
 
/** 
 * find the independence for variable X any Y given a set of other variables 
 * the significant level will be signif 
 * 
 * x is the name of first variable 
 * y is the name of second variable 
 * condition is the name array of condition variables, if it's empty, it will 
 * be null 
 * 
 * if x and y are independent given condition, it will return true. 
 */ 
 
public g2Result calcIndependence(String x, String y, String[] condition, double signif){ 
    int i,j; 
    double g2, df, prob; 
    boolean indep; 
    g2Result g2r; 

    //error check 
    if(isAnError(x,y,condition)) { 
      indep = false; 
      g2 = -1.0; 
      prob = -1.0; 
      df = -1.0; 
      } 
 
    int variable_index_length; 
    if (condition == null) 
      variable_index_length = 2; 
    else  
      variable_index_length = 2 + condition.length; 
 
    this.variable_index = new int[variable_index_length]; 
    this.signif = signif; 
 
    //set up variable_index[] 
    for (i=0; i<total_num_variable; i++){ 
      if(x.equals(varArray[i].name)) 
        this.variable_index[0] = i; 
      else if (y.equals(varArray[i].name)) 
        this.variable_index[1] = i; 
  } 
    for(i=2;i<variable_index_length; i++){ 
      for(j=0;j<total_num_variable;j++){ 
        if(condition[i-2].equals(varArray[j].name)) { 
          variable_index[i] = j;  
          break; 
        } 
      } 
    } 

    if (this.sample_size < 10*totalCell()){ 
 
       System.out.println("Sample size is less than 10 times the number of cells to be fitted, we assume the variables are conditionally dependent."); 
       indep = false; 
       g2 = -1.0; 
       prob = -1.0; 
       df = -1.0; 
       g2r = new g2Result(g2, prob, df, indep); 
       return g2r; 
    } 
 
    setupindices(); 
     
    fillincells(); 
 
    g2 = calcg2(); 
System.out.println("G2 test statistic: "+g2); 

    //calculate degrees of freedom 
    df = (varArray[variable_index[0]].size - 1) * (varArray[variable_index[1]].size - 1); 
    for(i = 2; i < variable_index.length; i++) 
      df = df * varArray[variable_index[i]].size; 
    df = df - this.num_zero_exp; 
    //System.out.println("the df is "+df);     
    if(df<1){ 
      System.out.println("the df is "+ df+ " which cannot be less than 1!"); 
      indep = false; 
      g2 = -1.0; 
      prob = -1.0; 
      df = -1.0; 
      g2r = new g2Result(g2, prob, df, indep); 
      return g2r; 
 
    } 
       
    prob = 1 - Probability.chisqCDF(g2, df); 
    //System.out.println("The p-value: "+prob); 
 
    if(prob > this.signif) 
      indep = true; 
    else indep = false; 
   
    g2r = new g2Result(g2, prob, df, indep); 
    return g2r; 
  } 
 
   
/** 
 * constructor for G2stat_test class 
 * input an instance of a Bayes network sample (bayesNetSample). 
 */ 
 
  public G2stat_test(bayesNetSample bns) { 
 
    String name; 
    int size ; 
    int[] cat; 
 
    this.sample_size = bns.sampleSize; 
    this.total_num_variable = bns.B.D.nvertices; 
 
    this.data = bns.cases; 
    this.varArray = new Variable[total_num_variable]; 
     
 
    for (int i = 0; i <total_num_variable; i++){ 
      //name 
      name = bns.B.D.vertexNames[i]; 
      //size 
      size = bns.B.nvalues[i]; 
      //categories 
      cat = new int[size]; 
      for(int j = 0; j < size; j++) { 
        cat[j] = j; 
        } 
      varArray[i] = new Variable(name, size, cat); 
      } 
     
  }//end of constructor 
 
/** 
 * Another constructor for G2stat_test class by Heiko Bailer 
 */ 
 
                   
  public G2stat_test(int currspots) 
  { 
    this.sample_size=rectangleFrame.sample_size; 
    this.total_num_variable=currspots; 
    this.data=rectangleFrame.data;  
    this.varArray = new Variable[total_num_variable]; 
    

System.out.println("\nConstructor: G2 test");

    for(int i=0;i<total_num_variable;i++)
    {
 varArray[i] = new Variable(rectangleFrame.varArray[i].name,rectangleFrame.varArray[i].size,rectangleFrame.varArray[i].category);
    }
  }//end of constructor 
 
} 
 
