//Copyright (C) 1997 Frank C. Wimberly.  All rights reserved.
class bayesNet {

  protected diGraph D;
  protected int nvalues[];
  //protected condDist dist[];
  protected float probs[][];
  protected int tableSize[];
  protected int pproduct[][];
  protected float updatedProbs[][];

  public bayesNet(diGraph D, int nvalues[]) 
  {
    this.D = D;
    this.nvalues = nvalues;
    //this.dist = new condDist[D.nvertices];
    this.probs = new float[D.nvertices][];
    this.tableSize = new int[D.nvertices];
    this.pproduct = new int[D.nvertices][D.nvertices];
    this.updatedProbs = new float[D.nvertices][];

    for(int i = 0; i < D.nvertices; i++) {
      updatedProbs[i] = new float[nvalues[i]];
      for(int j = 0; j < nvalues[i]; j++)
        updatedProbs[i][j] = (float) -1.0;
      }
  }

  public void defaultCondDist( )
  {
   
  for (int vertex = 0; vertex < D.nvertices; vertex++) {
    float prob;
    int product = 1;
    int parents = -1;
    for(int v = 0; v < D.nvertices; v++)
      if(D.adjacent(v, vertex)) {
        parents++;
        product = product*nvalues[v];
        pproduct[vertex][parents] = product;
        }
    int numValues = nvalues[vertex]*product;
    probs[vertex] = new float[numValues];
    tableSize[vertex] = numValues;
    // Set default probabilities
    prob = (float) 1.0/((float) nvalues[vertex]);
    for(int i = 0; i <numValues; i++) probs[vertex][i] = prob;
    }
  }

  public void randomConditionalDistribution(long seed) 
  {
   float rprobs[ ];
   float sum;
   int nv;

   nrRandom p = new nrRandom(seed);

   for(int vertex = 0; vertex < D.nvertices; vertex++) {
     nv = nvalues[vertex];
     int ncpv = tableSize[vertex]/nv;
     rprobs = new float[nv];
     for(int pc = 0; pc < ncpv; pc++) {
       float u = p.nrnextFloat( );
       rprobs[0] = u;
       sum = u;

       for(int i = 1; i < nv - 1; i++) {
         u = p.nrnextFloat( );
         rprobs[i] = ((float) 1.0 - sum)*u;
         sum += rprobs[i];
         }

       rprobs[nv - 1] = (float) 1.0 - sum;
       int istart = pc*nv;
       int iend = istart + nv;
       int j = 0;
       for(int i = istart; i < iend; i++) {
         probs[vertex][i] = rprobs[j];
         j++;
         }
       }
     }
  }


public int locateDistribution(int vertex, int parvals[]) 
  {
  int factor;
  int np = D.numParents[vertex];
  int start = 0;
  for(int p = 0; p < np; p++) {
    if( p == 0 ) factor = 1;
    else factor = pproduct[vertex][p-1];
    start += parvals[p]*factor;
    }
  start = start*nvalues[vertex];
  return start;
  }

public float conditionalProb(int vertex, int parvals[], int val) 
  { 
    int i;
    if(D.numParents[vertex] == 0) i = 0; else
      i = locateDistribution(vertex, parvals);
    return probs[vertex][i + val];
  }

public void storeConditionalProb(int vertex, int parvals[], int val, float p) 
  {
    int i = locateDistribution(vertex, parvals);
    probs[vertex][i + val] = p;
  }


public void estimateDistribution(bayesNetSample bns) 
  {
    int jj;
    boolean reading = true;
    int value[] = new int[D.nvertices];
    int eprobs[][];
    int vertex;

    eprobs = new int[D.nvertices][];
    for(vertex = 0; vertex < D.nvertices; vertex++) eprobs[vertex] = new int[tableSize[vertex]]
;


    for (int k = 0; k < bns.sampleSize; k++)  {
      for (int j = 0; j < D.nvertices; j++) value[j] = bns.cases[k][j];
      for(vertex = 0; vertex < D.nvertices; vertex++) {
        int np = D.numParents[vertex];
        int parvals[] = new int[np];
        for(int p = 0; p < np; p++) {
          int par = D.parent[vertex][p];
          parvals[p] = value[par];
          }
        int start = locateDistribution(vertex, parvals);
        eprobs[vertex][start + value[vertex]]++;
        }
      }
    //Put here the code for converting the eprobs table to the probs table
    for(vertex = 0; vertex < D.nvertices; vertex++) {
      int ncpv = tableSize[vertex]/nvalues[vertex]; //number of combinations of parent values
      for(int pc = 0; pc < ncpv; pc++) {
        int start = pc*nvalues[vertex];
        int sum = 0;
        for(int i = 0; i < nvalues[vertex]; i++) sum = sum + eprobs[vertex][start+i];
        for(int i = 0; i < nvalues[vertex]; i++) 
                  probs[vertex][start+i] = ((float) eprobs[vertex][start+i]) / ((float) sum);
        }
      }
  }

public void update(int evidence[]) 
{
  //assumes evidence[i] equals value of variable i if it is evidence; -1 otherwise.

  int varsToUpdate[] = new int[D.nvertices];
  int numToUpdate = 0;
  int tableSize = 1;
  int product[] = new int[D.nvertices];
  boolean added[] = new boolean[D.nvertices];
  int values[] = new int[D.nvertices];

  /* Debug print statements--Is the joint distribution correct?
  for(int j0 = 0; j0 < nvalues[0]; j0++) {
    values[0] = j0;
    for(int j1 = 0; j1 < nvalues[1]; j1++) {
      values[1] = j1;
      for(int j2 = 0; j2 < nvalues[2]; j2++) {
        values[2] = j2;
        for(int j3 = 0; j3 < nvalues[3]; j3++) {
          values[3] = j3;
          for(int j4 = 0; j4 < nvalues[4]; j4++) {
            values[4] = j4;
            float prob = jointProb(values);
     System.out.println("Row "+values[0]+values[1]+values[2]+values[3]+values[4]+" "+prob);
            }
          }
        }
      }
    }
  End of debug print statements */

  for(int i = 0; i < D.nvertices; i++)
    added[i] = false;

  for(int i = 0; i < D.nvertices; i++)
    if(evidence[i] >= 0) 
    for(int j = 0; j < D.nvertices; j++)
      if(existsTrek(i, j) && evidence[j] < 0 && !added[j]) {
        varsToUpdate[numToUpdate] = j;
        added[j] = true;
        System.out.println("Update " + D.vertexNames[j]);
        numToUpdate++;
        }
  System.out.println("Number to update = " + numToUpdate);
   
  product[0] = 1;
  for(int i = 1; i < D.nvertices; i++) {
    tableSize = tableSize*nvalues[i];
    product[i] = nvalues[i]*product[i-1];
    }
  tableSize = tableSize*nvalues[0];

//  createJointDist(numToUpdate, varsToUpdate, tableSize);

  //for(int v = 0; v < numToUpdate; v++) {
  //int var = varsToUpdate[v];
  for(int var = 0; var < D.nvertices; var++) {
    for(int v1 = 0; v1 < nvalues[var]; v1++) {
      //int values[] = new int[D.nvertices];

      //for each value of each variable in the to-be-updated list
      //sum1 = sum1 + joint(V = v1, E1=e1,...,Enev = eenv, X = x, Y = y, ... )
      //Letters without numerical subscripts are variables to be summed over
      //for each value of each variable in the to-be-updated list
      //sum2 = sum2 + joint(V = v, E1=e1,...,Enev = eenv, X = x, Y = y, ... )
      //Letters without numerical subscripts are variables to be summed over

      values[var] = v1;
      for(int i = 0; i < D.nvertices; i++)
        if(evidence[i] >= 0) values[i] = evidence[i];

      float sum1 = (float) 0.0;
      float sum2 = (float) 0.0;
      for (int row = 0; row < tableSize; row++) {
        calcValues(row, values, product);
        boolean in1 = true;
        boolean in2 = true;
        for(int i = 0; i < D.nvertices; i++) {
          if(values[var] != v1) in1 = false;
          if(evidence[i] >= 0 && evidence[i] != values[i]) {
            in1 = false;
            in2 = false;
            }
          }
        if (in2) sum2 += jointProb(values); //System.out.println("Sum2 row " + row);
        if (in1) sum1 += jointProb(values); //System.out.println("Sum1 row " + row);
        }

      //if evidence[i] >= 0) P(var = evidence[i] = 1.0) else P(var = v1) = sum1/sum2;
      float prob = sum1/sum2;
      if(var == 0) System.out.println("sum1 = " + sum1 + " sum2 = " + sum2);
      System.out.println("Variable " + D.vertexNames[var] + " value " + v1 + " Prob= " + prob);
      updatedProbs[var][v1] = prob; 
      }
    }
  }

public float updatedProb(int variable, int value)
{
  return updatedProbs[variable][value];
  }

public void createJointDist(int tableSize)
{
  int jointTable[] = new int[tableSize];  //nvertices x product of numbers of values
  }

public float jointProb(int values[])
{
  float jp = (float) 1.0;
  int parvals[] = new int[D.nvertices];

  for(int var=0; var < D.nvertices; var++) {
    if(D.numParents[var] == 0) jp = jp*conditionalProb(var, parvals, values[var]);
    else {
      for(int par = 0; par < D.numParents[var]; par++) 
        parvals[par] = values[D.parent[var][par]];
      jp = jp*conditionalProb(var, parvals, values[var]);
      }
    }
  return jp;
}

public int calcIndex(int values[], int product[]) 
{ 
  int index = 0;
  for(int var = 0; var < D.nvertices; var++)
    index += values[var]*product[D.nvertices - var - 1];
  return index;
  }

public void calcValues(int index, int values[], int product[]) 
{
  int rem = index;
  for(int i = 0; i < D.nvertices; i++) {
    values[i] = rem / product[D.nvertices - i - 1];
    rem = rem % product[D.nvertices - i - 1];
    }
  }

public boolean existsTrek(int x, int y)
{
  if(D.isPath(x, y) || D.isPath(y, x)) return true;

  for(int i = 0; i < D.nvertices; i++)
    if((i != x) && (i != y) && D.isPath(i, x) && D.isPath(i, y)) return true;

  return false;
  }

public bayesNet copy( )
{
  bayesNet newB;
  diGraph newD;
  int newEdges[][] = new int[D.nvertices][D.nvertices];

  for(int i = 0; i < D.nvertices; i++) 
    for(int j = 0; j < D.nvertices; j++)
      newEdges[i][j] = D.edgeMatrix[i][j];

  newD = new diGraph(D.nvertices, newEdges, D.name, D.vertexNames);
  newB = new bayesNet(newD, nvalues);
  newD.setupTiers( );
  newD.findParents( );
  newB.defaultCondDist( );

  for(int vertex = 0; vertex < newD.nvertices; vertex++) {
    int np = newD.getNumParents(vertex);
    newB.probs[vertex] = new float[tableSize[vertex]];
    for(int i = 0; i < tableSize[vertex]; i++)
      newB.probs[vertex][i] = probs[vertex][i];
    }
  return newB;
  }
}
