(*-----------#unix-version---------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "util.h"
#include "io.h"
#include "calctet.h"
#include "calctet1.h"
#include "sys.h"
#include "eqs.h"
#include "part.h"
#include "scal.h"
#include "mon.h"
#include "rsutil.h"
#include "thomas.h"

type elementptr = ^elementlist;
     elementlist = record
			next: elementptr;
			left: vertextype;
			middle: vertextype;
			right: vertextype;
		     end;

type elempairptr = ^elempairlist;
     elempairlist = record
                     next: elempairptr;
                     first: elementptr;
                     second: elementptr;
                 end;

type elemelempairptr = ^elemelempairlist;
      elemelempairlist = record
                          next: elemelempairptr;
                          first: elementptr;
                          second: elempairptr;
                         end;

function bigger(a,b:integer):integer;
begin
if a >= b then bigger:= a else bigger := b;
end;

function smaller(a,b:integer):integer;
begin
if a >= b then smaller:= b else smaller := a;
end;

procedure calcapath(var adj1,adj2: adjtype);
var j,k,i: integer;
   
begin (* calcpath *)
   for i := 1 to numob do 
   for j := 1 to numob do
   adj2^[i,j] := adj1^[i,j];
   for i := 1 to numob do
   adj2^[i,i] := 1;
   for k := 1 to numob do 
   for i := 1 to numob do 
   for j := 1 to numob do        
   if (adj2^[i,k] <> notthere) and
      (adj2^[k,j] <> notthere) then
      adj2^[i,j] := 1;
end; (* Calcdpath *)

procedure copyadjtograph(var graph: adjtype);
var i,j: vertextype;

begin
   for i := 1 to numob do
   for j := 1 to numob do
   graph^[i,j] := adjacency^[i,j];
end;

procedure showgraph(var adj1: adjtype; var tofile: text; literal: boolean);
(*showgraph prints an (adjacency) matrix                                     *)
(*if literal is true then the procedure gives the actual values in the matrix*)
(*if literal is false then the values are displayed mod 2                    *)
var j,k: vertextype;

begin
   twrite(tofile,'       (second index)|');
   twriteeoln(tofile);
   twrite(tofile,'        |');
   for j := 1 to numob do
   begin
   pwritename(j,tofile,4);
{   twrite(tofile, '  |');}
   end;
   twriteeoln(tofile);
   for j := 1 to numob do
   begin
        twrite(tofile,'    |');
        pwritename(j,tofile,1);
   	for k := 1 to numob do
  	  begin
		if (literal) then
    		tpwriteint(tofile,adj1^[j,k],4,true)
    		else tpwriteint(tofile,(adj1^[j,k] mod 2),4,true);   
	  end;
        twriteeoln(tofile);
   end;
end;

   
procedure lslice(*var adj1: adjtype; v1: vertextype;var temp: vertexset*);
{adj1 - the input graph; v1 - a vertex in that graph
 temp - the set of vertices which is output
 lslice gives the set of vertices (x | adj1^[x,v1] is there)
 If supplied with an ordinary graph matrix as input, lslice will give the set
 of children of v1 as output}
var i: vertextype;

begin
temp := [] ;
for i :=1 to numob do
    if adj1^[i,v1]<>notthere then temp := (temp + [i]);
end;


procedure rslice(*var adj1: adjtype; v1: vertextype;var temp: vertexset*);
{adj1 - the input graph; v1 - a vertex in that graph
 temp - the set of vertices which is output
 rslice gives the set of vertices (x | adj1^[v1,x] is there)
 If supplied with an ordinary graph matrix as input, rslice will give the set
 of parents of v1 as output}
var i: vertextype; 

begin
temp := [] ;
for i :=1 to numob do
    if adj1^[v1,i]<>notthere then temp := (temp + [i]);
end;


procedure testlslice(var adj1:adjtype);
var j,i: vertextype;
    par1,par2,temp1,temp2,temp3,temp4: vertexset;
begin
for j := 1 to numob do
  begin
    lslice(adj1,j,temp1);
    twrite(output,'temp1 <|');tpwriteint(output,j,1,true);
    twrite(output,' > [|');
    for i := 1 to numob do
    if i in temp1 then tpwriteint(output,i,2,true);
    twriteln(output,'] |');
  end;
end;

procedure virtualadjs(var adj1,adj2,adj3: adjtype); 
{adj1 is the input matrix of real edges			        }
{adj2 is the input matrix of ancestor relations			}
{adj3 is the output matrix of adjacencies (symmetric)		}

var i,j,k: vertextype;
    par1,par2,temp1,temp2,temp3,temp4: vertexset;
begin
for j := 1 to numob do
 for k := j to numob do
  if j <> k then
         begin
         rslice(adj1,j,temp1);
         rslice(adj1,k,temp2);
         lslice(adj2,j,temp3);
         lslice(adj2,k,temp4);
             if (adj1^[j,k]<>notthere) or (adj1^[k,j]<>notthere) or   
             ((temp1 * temp2) * (temp3 + temp4) <> []) then
                               begin  
                               adj3^[j,k] := 1;
                               adj3^[k,j] := 1; 
                               end
                            else 
                               begin
                               adj3^[j,k] := 2;
                               adj3^[k,j] := 2; 
                               end     
         end
  else 
     begin
        adj3^[j,k] := 2; {We set the diagonal elements}
        adj3^[k,j] := 2; {to 'notthere' as this simplifies the}
     end;                {procedure which looks for m.e. conds}
end;

procedure initelemlist(var elemlist:elementptr);

begin
   new(elemlist);
   elemlist^.left := 0;
   elemlist^.right := 0;
   elemlist^.middle := 0;
   elemlist^.next := nil;
end;

procedure initelempairlist(var elemplist:elempairptr);

begin
   new(elemplist);
   elemplist^.first := nil;
   elemplist^.second := nil;
   elemplist^.next := nil;
end;

procedure initelemelempairlist(var elemelemplist:elemelempairptr);

begin
  new(elemelemplist);
  elemelemplist^.first := nil;
  elemelemplist^.second := nil;
  elemelemplist^.next := nil;
end;

procedure writelemlist(var elemlist:elementptr; lft,midle,rght: vertextype);

begin
  elemlist^.left := lft;
  elemlist^.middle := midle;
  elemlist^.right := rght;
  initelemlist(elemlist^.next);
  elemlist:=elemlist^.next;

end;

procedure writelempairlist(var elemplist:elempairptr; fst,secnd:elementptr);

begin
  elemplist^.first := fst;
  elemplist^.second := secnd;
  initelempairlist(elemplist^.next);
  elemplist := elemplist^.next;
end;

procedure writelemelempairlist(var elemelemplist:elemelempairptr; 
                               var fst:elementptr; var secnd:elempairptr);

begin
  elemelemplist^.first := fst;
  elemelemplist^.second := secnd;
  initelemelempairlist(elemelemplist^.next);
  elemelemplist := elemelemplist^.next;
end;

   
procedure classifyelements(var adj1,adj2,adj3: adjtype; 
                var cons,pmecons,impnoncons,pernoncons: elementptr);
{adj1 is the input set of real adjacencies
 adj2 is the input set of ancestral relations
 adj3 is the input set of virtual adjacencies
 cons is the output list of conductors
 pmecons is the output list of potentially m.e. conductors
 impnoncons is the output list of imperfect non-conductors
 pernoncons is the output list of perfect non-conductors}

var i,j,k: vertextype;
var tempi,tempj,tempk: vertexset;
var curcons,curpmecons,curimpnoncons,curpernoncons: elementptr


begin
  initelemlist(pernoncons);initelemlist(impnoncons);
  initelemlist(cons);initelemlist(pmecons);

  curpernoncons:=pernoncons; curimpnoncons:=impnoncons;
  curpmecons:=pmecons; curcons:=cons;
for i:=1 to numob do
     for j:=i to numob do
          if i<>j then
              for k := 1 to numob do
               if (i<>k) and (j<>k) then 
                    if ((adj3^[i,k]<>notthere) and (adj3^[k,j]<>notthere)
                    and (adj3^[i,j] = notthere)) then
                         if ((adj2^[k,i]<>notthere) or (adj2^[k,j]<>notthere))
                         then begin 
                             writelemlist(curcons,i,k,j);
                             if ((adj2^[k,i]=notthere) and (adj2^[j,k]<>notthere)) 
                              then writelemlist(curpmecons,i,k,j);
                             if ((adj2^[k,j]=notthere) and (adj2^[i,k]<>notthere))
                              then writelemlist(curpmecons,j,k,i);
                              end
                         else begin
                             rslice(adj1,i,tempi);
                             rslice(adj1,j,tempj);
			     lslice(adj2,k,tempk);
                             if ((tempi*tempj)*tempk)<>[] then
                               writelemlist(curpernoncons,i,k,j)
                             else writelemlist (curimpnoncons,i,k,j);
			     end;
end;




procedure formsearchmatrix(var adj1,adj2,adj3:adjtype;
                                var pmecon1,pmecon2:elementptr);
{adj1 is the input ancestor matrix
 adj2 is the input adjacency matrix
 adj3 is the output m.e. search matrix
 pmecon1,pmecon2 are pointers to the two pme conductors whose
 mutual exclusivity the matrix is designed to test
 }

var i,j,v1,v2,v3,w1,w2,w3:vertextype;

begin
v1:=pmecon1^.left;v2:=pmecon1^.middle;v3:=pmecon1^.right;
w1:=pmecon2^.left;w2:=pmecon2^.middle;w3:=pmecon2^.right;

for i:= 1 to max do
    for j:= i to max do
      if i<>j then
          if  (adj1^[i,j]<>notthere) and (adj1^[j,i]<>notthere)
           and (adj2^[i,j]<>notthere)
           and ((adj2^[v1,i]=notthere) or (i=v2)) 
           and ((adj2^[v1,j]=notthere) or (j=v2))
           and ((adj2^[v2,i]=notthere) or (i=v3)) 
           and ((adj2^[v2,j]=notthere) or (j=v3))
           and ((adj2^[w1,i]=notthere) or (i=w2)) 
           and ((adj2^[w1,j]=notthere) or (j=w2))
           and ((adj2^[w2,i]=notthere) or (i=w3)) 
           and ((adj2^[w2,j]=notthere) or (j=w3))
                then begin adj3^[i,j]:=1; adj3^[j,i]:=1; end
                else begin adj3^[i,j]:=2; adj3^[j,i]:=2; end
     
end;

procedure findmecons(var adj1,adj2:adjtype; 
                        var pmeconlist:elementptr; var mepairlist:elempairptr);
{adj1 is the input matrix of ancestral relations
 adj2 is the input matrix of adjacencies
 pmeconlist is the list of pme conductors
 mepairlist is the list of pairs of mutually exclusive conductors}

var curptr1,curptr2:elementptr;
var mematrix,amematrix:adjtype;
var curptrme:elempairptr;
begin
tnewadj(mematrix,fatal);
tnewadj(amematrix,fatal);
curptr1:=pmeconlist;
curptr2:=pmeconlist;
initelempairlist(mepairlist);
curptrme:=mepairlist;
while (curptr1^.next<>nil) do
   begin 
   while (curptr2^.next<>nil) do 
       begin
       if (curptr1<>curptr2)
           and (curptr1^.left<>curptr2^.left) 
           and (curptr1^.middle<>curptr2^.middle)
           and (adj2^[(curptr1^.left),(curptr2^.left)]=notthere)
           then begin
              formsearchmatrix(adj1,adj2,mematrix,curptr1,curptr2);
              calcapath(mematrix,amematrix);
	       if (amematrix^[(curptr1^.middle),(curptr2^.middle)]<>notthere)
                 then begin
                      writelempairlist(curptrme,curptr1,curptr2);
        {*****         twriteeoln(output);
                       writegraph(mematrix,output,false,false);
                       showgraph(mematrix,output,false);
                       showgraph(amematrix,output,false);
                       twrite(output,'***   |');
                       printelem(output,curptr1);
                       twrite(output,' and |');
                       twrite(output,'<|');
                       pwritename(curptr2^.right,output,2);twrite(output,',|');
                       pwritename(curptr2^.middle,output,2);twrite(output,',|');
                       pwritename(curptr2^.left,output,2);twrite(output,',|');
                       twrite(output,'> are m.e.! ***|');******}
                      end; 
            end;
       curptr2:=curptr2^.next;
       end;
    
    curptr1 := curptr1^.next;
    curptr2 := curptr1
    end;
dispose(mematrix);
dispose(amematrix);
end;

procedure impimpdiamond (var adj1: adjtype; var impnonconlist:elementptr;
                                      var impimplist:elempairptr);

{adj1 is the input matrix of ancestral relations
 impnonconlist is the input list of imperfect non-conductors
 impimp list is the list of ordered pairs of imperfect non-conductors
 such that the middle element of the first is the ancestor of the middle
 element of the second}

var curptr1,curptr2:elementptr;
var curptrimpimp:elempairptr;

begin
initelempairlist(impimplist);
curptr1 := impnonconlist;
curptr2 := impnonconlist;
curptrimpimp := impimplist;

while (curptr1^.next<>nil) do
   begin 
   while (curptr2^.next<>nil) do 
       begin
       if (curptr1<>curptr2)
          and (curptr1^.left = curptr2^.left)
          and (curptr1^.right = curptr2^.right)
          and (adj1^[curptr1^.middle,curptr2^.middle]<>notthere)
          then
           writelempairlist(curptrimpimp,curptr1,curptr2);
       curptr2:=curptr2^.next;
       end;
    curptr1 := curptr1^.next;
    curptr2 := impnonconlist
    end;
end;

procedure impmediamond (var adj1: adjtype; var impnonconlist:elementptr;
                      var meconlist: elempairptr; var impmelist:elemelempairptr);
{adj1 is the input matrix of ancestral relations
 impnonconlist is the input list of imperfect non-conductors
 meconlist is the input list of mutually exclusive conductors
 impmelist is the output list of me conductor pairs and imperfect 
 non-conductors which are such that the middle of the imp. non-conductor
 is the descendant of the middles of (both) the m.e. conductors}

var curptrimp:elementptr;
var curptrme:elempairptr;
var curptrimpme:elemelempairptr;

begin
initelemelempairlist(impmelist);
curptrimp := impnonconlist;
curptrme := meconlist;
curptrimpme := impmelist;

while (curptrimp^.next<>nil) do
   begin 
   while (curptrme^.next<>nil) do 
       begin
       if (curptrme^.first^.left=curptrimp^.left)
          and (curptrme^.second^.left=curptrimp^.right)
          and (curptrme^.first^.middle<>curptrimp^.middle)
          and (curptrme^.second^.middle<>curptrimp^.middle)
          and (adj1^[curptrme^.second^.middle,curptrimp^.middle]<>notthere)
      then
           writelemelempairlist(curptrimpme,curptrimp,curptrme);
       curptrme:=curptrme^.next;
       end;
    curptrimp := curptrimp^.next;
    curptrme := meconlist
    end;
end;





procedure printelem(var tofile:text; var curptr:elementptr);

begin
    twrite(tofile,'<|');
    pwritename(curptr^.left,tofile,2);twrite(tofile,',|');
    pwritename(curptr^.middle,tofile,2);twrite(tofile,',|');
    pwritename(curptr^.right,tofile,2);twrite(tofile,',|');
    twrite(tofile,'> |');
end;



procedure printelemlist(var tofile: text; var temp:elementptr);

var curptr:elementptr;
var written:integer;
begin

curptr := temp;
written:=0;
while (curptr^.next)<>nil do
       begin
 	  printelem(tofile,curptr);
	  written:=written+1;
          if (written mod 4) = 0 then twriteeoln(tofile);
          curptr:=curptr^.next;
       end;

end;
        
procedure printclassif(var tofile: text;
                      var cons,pmecons,impnoncons,pernoncons:elementptr);
begin
twrite(tofile,'3. Unshielded Conductors:|');twriteeoln(tofile);
printelemlist(tofile,cons);
twriteeoln(tofile);
twrite(tofile,'4. Potentially Mutually Exclusive Unshielded Conductors:|');twriteeoln(tofile);
printelemlist(tofile,pmecons);
twriteeoln(tofile);
twrite(tofile,'5. Unshielded Perfect Non-Conductors:|');twriteeoln(tofile);
printelemlist(tofile,pernoncons);
twriteeoln(tofile);
twrite(tofile,'6. Unshielded Imperfect Non-Conductors:|');twriteeoln(tofile);
printelemlist(tofile,impnoncons);
twriteeoln(tofile);twriteeoln(tofile);
end;

procedure printmeconpair(var tofile: text; var first,second: elementptr);

begin
twrite(tofile,'{|');
printelem(tofile,first);
twrite(tofile,'... |');
twrite(tofile,'<|');
pwritename(second^.right,tofile,2);twrite(tofile,',|');
pwritename(second^.middle,tofile,2);twrite(tofile,',|');
pwritename(second^.left,tofile,2);twrite(tofile,',|');
twrite(tofile,'>}  |');
end;

procedure printmecons(var tofile: text; var mecons:elempairptr);

var curptr:elempairptr
var written: integer

begin
written := 0;
curptr := mecons;
twrite(tofile,'7. Pairs of Mutually Exclusive Unshielded Conductors:|');
twriteeoln(tofile);
twrite(tofile,'(Oriented as they lie on some uncovered itinerary)|');
twriteeoln(tofile);
while curptr^.next<>nil do
     begin
       printmeconpair(tofile,curptr^.first,curptr^.second);
       written := written+1;
       if (written mod 2) = 0 then twriteeoln(tofile);
       curptr:=curptr^.next
     end;
end;

procedure printimpimp(var tofile: text; var first,second:elementptr);

begin
printelem(tofile,second);twrite(tofile,'  |');printelem(tofile,first);
twrite(tofile,' - |');pwritename(second^.middle,tofile,2);
twrite(tofile, ' is a descendant of |');
pwritename(first^.middle,tofile,2);
twriteeoln(tofile);
end;

procedure printimpimplist(var tofile: text; var impimplist:elempairptr);
var curptr:elempairptr

begin
curptr := impimplist;
twrite(tofile, '8. Pairs of Unshielded Imperfect Non-conductors:|');
twriteeoln(tofile);
twrite(tofile, '(between the same endpoints)|');
twriteeoln(tofile);
while curptr^.next<>nil do
     begin
       printimpimp(tofile,curptr^.first,curptr^.second);
       curptr:=curptr^.next
       end;
end;

procedure printimpme(var tofile: text; first:elementptr; var second:elempairptr);

begin
printmeconpair(tofile,second^.first,second^.second);twrite(tofile,'  |');
printelem(tofile,first);
twrite(tofile,' - |');pwritename(first^.middle,tofile,2);
twrite(tofile, ' is a descendant of |');
pwritename(second^.first^.middle,tofile,2);twrite(tofile, ' and |');
pwritename(second^.second^.middle,tofile,2);
twriteeoln(tofile);
end;

procedure printimpmelist(var tofile: text; var impmelist:elemelempairptr);
var curptr:elemelempairptr

begin
curptr := impmelist;
twrite(tofile, '9. Unshielded Imperfect Non-conductors and pairs of unshielded m.e. conductors:|');
twriteeoln(tofile);
twrite(tofile, '(between the same endpoints)|');
twriteeoln(tofile);
while curptr^.next<>nil do
     begin
       printimpme(tofile,curptr^.first,curptr^.second);
       curptr:=curptr^.next
       end;
end;


procedure makeequivclass(var graph,vgraph:adjtype;
                 var impnonconlist,pernonconlist,conlist,pmeconlist:elementptr;
                 var mepairlist,impimplist:elempairptr;
                 var impmelist:elemelempairptr);
{graph is the input matrix
 vgraph is the output (real/virtual) adjacency matrix
 impnonconlist is the list of imperfect non-conductors
 pernonconlist is the list of perfect non-conductors
 mepairlist is the list of me conductors
 impimplist is the list of imperfect non-cons between the same endpts
 impmelist is the list of imperfect non-cons and m.e. conductors
 between the same endpts}


var agraph:adjtype;

begin
   tnewadj(agraph,fatal);
   calcapath(graph,agraph);
   virtualadjs(graph,agraph,vgraph);
   classifyelements(graph,agraph,vgraph, 
                conlist,pmeconlist,impnonconlist,pernonconlist);
   findmecons(agraph,vgraph,pmeconlist,mepairlist);
   impimpdiamond(agraph,impnonconlist,impimplist);
   impmediamond(agraph,impnonconlist,mepairlist,impmelist);  
   dispose(agraph);
end;


procedure printnoteq(var tofile:text);

begin
       twrite(tofile,'The graphs are not equivalent.|');
      twriteeoln(tofile);
end;

function same(var tmp1,tmp2:elementptr):boolean;

begin

if ((tmp1^.left=tmp2^.left)
      and (tmp1^.right=tmp2^.right)
      and (tmp1^.middle=tmp2^.middle)) 
      then same:=true
      else same:=false;
end;

function samepair(var tmp1,tmp2:elempairptr):boolean;
label 747;
begin
if (tmp1^.next=nil) or (tmp2^.next=nil) then 
      begin samepair:=false; goto 747;end;

if ((tmp1^.first^.left=tmp2^.first^.left)
      and (tmp1^.first^.right=tmp2^.first^.right)
      and (tmp1^.first^.middle=tmp2^.first^.middle)
      and (tmp1^.second^.left=tmp2^.second^.left)
      and (tmp1^.second^.right=tmp2^.second^.right)
      and (tmp1^.second^.middle=tmp2^.second^.middle))
   then samepair:=true
   else samepair:=false;

747:end;


procedure comparelist(var tmp1,tmp2,elem1,elem2 :elementptr; var eqal:boolean);

var curptr1,curptr2:elementptr;

begin
curptr1:=tmp1;
curptr2:=tmp2;

while ((curptr1^.next<>nil) and  same(curptr1,curptr2)) do
           begin curptr1:=curptr1^.next;curptr2:=curptr2^.next end; 
{Note if curptr2^.next=nil and curptr1^.next<>nil then 
  curptr1^.right<>curptr2^.right, since if curptr2^.next=nil
  then curptr2^.right=0 by intitialization}
   
if not same(curptr1,curptr2)
      then begin elem1:=curptr1; elem2:=curptr2; eqal:=false end;
end;

procedure comparelempairlist(var tmp1,tmp2:elempairptr;
                          var elem1,elem2 :elempairptr; var eqal:boolean);

var curptr1,curptr2:elempairptr;

label 683;

begin
curptr1:=tmp1;
curptr2:=tmp2;

while ((curptr1^.next<>nil) or (curptr2^.next<>nil)) do 
   begin
   if not(samepair(curptr1,curptr2)) then
        begin elem1:=curptr1; elem2:=curptr2; eqal:=false; goto 683 end; 
   curptr1:=curptr1^.next;curptr2:=curptr2^.next
   end; 

683:end;

procedure comparelemelempairlist(var tmp1,tmp2:elemelempairptr;
                               var elem1,elem2 :elemelempairptr; var eqal:boolean);
label 710;

var curptr1,curptr2:elemelempairptr;

begin
curptr1:=tmp1;
curptr2:=tmp2;
while ((curptr1^.next<>nil) or (curptr2^.next<>nil)) do
   begin
   if ((curptr1^.next=nil) or (curptr2^.next=nil)) then
     begin elem1:=curptr1; elem2:=curptr2; eqal:=false; goto 710 end 
   else if not(samepair(curptr1^.second,curptr2^.second)
               and same(curptr1^.first,curptr2^.first) then
       begin elem1:=curptr1; elem2:=curptr2; eqal:=false; goto 710 end; 
   curptr1:=curptr1^.next;curptr2:=curptr2^.next
   end;
710:end;


procedure getfirstelem(var con1,con2:elementptr);
{This procedure looks at two elements and then works
out which one should have been classified *first* by the
program. It then returns the element which occurs first
in con1.}

begin
if (not(con2^.left = 0) and ((con1^.left > con2^.left)
           or ((con1^.left = con2^.left) and (con1^.right > con2^.right))
           or ((con1^.left = con2^.left) and (con1^.right = con2^.right)
                 and (con1^.middle > con2^.middle))
           or (con1^.left=0))) then con1:=con2;
end;



procedure getfirstinpair(var pair1,pair2:elempairptr);

{This procedure looks at two m.e. conductor pairs and works out 
 which one's existence should have been checked for *first*
 by the program. It then returns the pair which occurs first
 in this ordering in pair1}
label 747;
var x1,x2,x3,y1,y2,y3:vertextype;
var u1,u2,u3,v1,v2,v3:vertextype;

begin
if (pair1^.first=nil) then pair1:=pair2; goto 747;
if (pair2^.first=nil) then goto 747;

     x1 := bigger(pair1^.first^.right,pair1^.first^.left);
     x2 := smaller(pair1^.first^.right,pair1^.first^.left);
     x3 := pair1^.first^.middle;
     y1 := bigger(pair1^.second^.right,pair1^.second^.left);
     y2 := smaller(pair1^.second^.right,pair1^.second^.left);
     y3 := pair1^.first^.middle;
     u1 := bigger(pair2^.first^.right,pair2^.first^.left);
     u2 := smaller(pair2^.first^.right,pair2^.first^.left);
     u3 := pair2^.first^.middle;
     v1 := bigger(pair2^.second^.right,pair2^.second^.left);
     v2 := smaller(pair2^.second^.right,pair2^.second^.left);
     v3 := pair2^.first^.middle;

if ((x1>u1) or ((x1=u1) and ((x2>u2) or ((x2=u2) and ((x3>u3)
        or ((x3=u3) and ((y1>v1) or ((y1=v1) and ((y2>v2) 
        or ((y2=v2) and (y3>v3))))))))))) 
        then pair1:=pair2;
747:end;

procedure compare(var tofile:text;
                  var vgraph1,vgraph2:adjtype;
                  var impnonconlist1,impnonconlist2:elementptr;
                  var pernonconlist1,pernonconlist2:elementptr;
                  var conlist1,conlist2:elementptr;
                  var mepairlist1,mepairlist2:elempairptr;
                  var impimplist1,impimplist2:elempairptr;
                  var impmelist1,impmelist2:elemelempairptr);
label 768;
var i,j,k:integer;
var eqal:boolean;
var con1,con2:elementptr;
var pair1,pair2:elempairptr;
var impme1,impme2:elemelempairptr;
         
begin
eqal:=true;
for i:= 1 to max do
  for j:= 1 to max do
    if (vgraph1^[i,j]<>vgraph2^[i,j]) then 
      begin
         if eqal then
            begin  printnoteq(output);printnoteq(tofile);end;   
         twrite(output,'Difference in adjacencies.|');         
         twrite(tofile,'Difference in adjacencies between: |');
         pwritename(i,tofile,2);twrite(tofile,' and |');
         pwritename(j,tofile,2);
         twriteeoln(tofile);
         eqal:=false;goto 768;         
      end;

if eqal then begin
    twrite(output,'Same adjacencies.|');
    twriteeoln(output);
    twrite(tofile,'Same adjacencies.|');
    twriteeoln(tofile)
    end
    else goto 768;
    

comparelist(conlist1,conlist2,con1,con2,eqal);
if not eqal then begin
        printnoteq(output);printnoteq(tofile);
        twrite(tofile,'Different unshielded conductors.|');  
        twrite(output,'Different unshielded conductors.|');
	twriteeoln(tofile);
        twrite(tofile,'Unshielded Conductor: |');   
        getfirstelem(con1,con2);
        printelem(tofile,con1);
        twriteeoln(tofile); 
        if (con1=con2) then twrite(tofile,'occurs in G2, but not in G1.|')
                       else twrite(tofile,'occurs in G1, but not in G2.|'); 
        twriteeoln(tofile); 
        goto 768; end      
   else begin 
        twrite(output,'Same Unshielded Conductors (if any).|');
        twriteeoln(output); 
        end;

comparelist(impnonconlist1,impnonconlist2,con1,con2,eqal);
if not eqal then begin
        printnoteq(output);printnoteq(tofile);
        twrite(tofile,'Different unshielded imperfect non-conductors.|');   
        twrite(output,'Different unshielded imperfect non-conductors.|');
        twriteeoln(tofile);  
        twrite(tofile,'Unshielded Imperfect Non-Conductor: |');
        getfirstelem(con1,con2);
        printelem(tofile,con1);
        twriteeoln(tofile); 
        if (con1=con2) then twrite(tofile,'occurs in G2, but not in G1.|')
                       else twrite(tofile,'occurs in G1, but not in G2.|'); 
        twriteeoln(tofile); 
        goto 768; end      
   else begin  
        twrite(output,'Same Unshielded Imperfect Non-Conductors (if any).|'); 
        twriteeoln(output);
        twrite(output,'Same Unshielded Perfect Non-Conductors (if any).|'); 
        twriteeoln(output);end;

{The next section checks that the two graphs have the same perfect non-conductors
 which should be automatically entailed by the fact that at this stage we know that
 they have the same adjacencies, conductors and imperfect non-conductors, hence
 it can probably be safely omitted.}
comparelist(pernonconlist1,pernonconlist2,con1,con2,eqal);
if not eqal then begin
        printnoteq(output);printnoteq(tofile);
        twrite(tofile,'This should be a logical impossibility!|');   
        twrite(output,'This should be a logical impossibility!.|');
        twriteeoln(tofile);
        twriteeoln(output);
        twrite(tofile,'Same adjacencies, conductors and imperfect non-conductors,|');
        twriteeoln(tofile);
        twrite(tofile,'should entail same perfect non-conductors.|');   
        twrite(output,'Same adjacencies, conductors and imperfect non-conductors,');
        twriteeoln(output);
        twrite(output,'should entail same perfect non-conductors.|');          
        twriteeoln(output);
        twriteeoln(tofile);
        twrite(tofile,'Different unshielded perfect non-conductors.|');   
        twrite(output,'Different unshielded perfect non-conductors.|');
        twriteeoln(tofile);
        twriteeoln(output);
        twrite(tofile,'Unshielded Perfect Non-Conductor: |');
        getfirstelem(con1,con2);
        printelem(tofile,con1);
        twriteeoln(tofile); 
        if (con1=con2) then twrite(tofile,'occurs in G2, but not in G1.|')
                       else twrite(tofile,'occurs in G1, but not in G2.|'); 
        twriteeoln(tofile); 
        goto 768; end;      
   

comparelempairlist(mepairlist1,mepairlist2,pair1,pair2,eqal);
if not eqal then begin
        printnoteq(output);printnoteq(tofile);
        twrite(tofile,'Different pairs of mutually exclusive (unshielded) conductors.|');   
        twrite(output,'Different pairs of mutually exclusive (unshielded) conductors.|');
        twriteeoln(tofile);
        twrite(tofile,'Mutually Exclusive (Unshielded) Conductor Pair: |');
        twriteeoln(tofile);
        getfirstinpair(pair1,pair2);
        printmeconpair(tofile,pair1^.first,pair1^.second);
        twriteeoln(tofile); 
        if (pair1=pair2) then twrite(tofile,'occurs in G2, but not in G1.|')
                       else twrite(tofile,'occurs in G1, but not in G2.|'); 
        twriteeoln(tofile); 
        goto 768; end      
   else begin
        twrite(output,'Same Pairs of Mutually Exclusive (Unshielded) Conductors (if any).|'); 
        twriteeoln(output);
        twrite(output,'Same Pairs of Mutually Exclusive (Unshielded) Conductors (if any).|'); 
        twriteeoln(output);end;

comparelempairlist(impimplist1,impimplist2,pair1,pair2,eqal);
if not eqal then begin
        printnoteq(output);printnoteq(tofile);
        twrite(tofile,'Different ancestor relations between unshielded imperfect non-conductors,|');
        twrite(output,'Different ancestor relations between unshielded imperfect non-conductors,|');
        twriteeoln(tofile);
        twriteeoln(output);
        twrite(tofile,'between the same endpoints,|');   
        twrite(output,'between the same endpoints.|');   
        if pair1^.first=nil then begin
                 printimpimp(tofile,pair2^.first,pair2^.second);
                 twriteeoln(tofile);
                 twrite(tofile, 'in G2, but this is not the case in G1.|');
                 goto 768; end;
        if pair2^.first=nil then begin
                 printimpimp(tofile,pair1^.first,pair1^.second);
                 twriteeoln(tofile);
                 twrite(tofile, 'in G1, but this is not the case in G2.|');
                 goto 768; end;
        if same(pair1^.first,pair2^.first) 
            then begin
                 getfirstelem(pair1^.second,pair2^.second);
                 printimpimp(tofile,pair1^.first,pair1^.second);
                 twriteeoln(tofile);
                 if (pair1^.second=pair2^.second) then 
                            twrite(tofile,'in G2, but this is not the case in G1.|')
                       else twrite(tofile,'in G1, but this is not the case in G2.|'); 
                 goto 768;end
            else begin
                 getfirstelem(pair1^.first,pair2^.first);
                 if (pair1^.first=pair2^.first) then 
                       begin
                       printimpimp(tofile,pair2^.first,pair2^.second);
                       twriteeoln(tofile);                 
                       twrite(tofile,'in G2, but this is not the case in G1.|');
                       goto 768 end
                  else begin
                       printimpimp(tofile,pair1^.first,pair1^.second);
                       twriteeoln(tofile);
                       twrite(tofile,'in G1, but this is not the case in G2.|'); 
                 goto 768;end;
                 end;
            end
   else begin
        twrite(tofile,'Same ancestor relations between unshielded imperfect non-conductors,|');
        twrite(output,'Same ancestor relations between unshielded imperfect non-conductors,|');
        twriteeoln(tofile);
        twriteeoln(output);
        twrite(tofile,'between the same endpoints, (if any)|');   
        twrite(output,'between the same endpoints, (if any)|');   
        twriteeoln(tofile);
        twriteeoln(output); end;

comparelemelempairlist(impmelist1,impmelist2,impme1,impme2,eqal);
if not eqal then begin
        printnoteq(output);printnoteq(tofile);
        twrite(tofile,'Different ancestor relations between unshielded imperfect non-conductors,|');
        twrite(output,'Different ancestor relations between unshielded imperfect non-conductors,|');
        twriteeoln(output);
        twriteeoln(tofile);
        twrite(tofile,'and (unshielded) m.e. conductors between the same endpoints,|');   
        twrite(output,'and (unshielded) m.e. conductors between the same endpoints.|');   
        twriteeoln(output);
        twrite(output,'This is a pretty bizarre event - I know of no actual examples.|');
        twriteeoln(output);
        twrite(output,'of two graphs which are the same in every other respect except|');
        twriteeoln(output);
        twrite(output,'this one!!|');
        twriteeoln(output);
        if impme1^.first=nil then begin
                 printimpme(tofile,impme2^.first,impme2^.second);
                 twriteeoln(tofile);
                 twrite(tofile, 'in G2, but this is not the case in G1.|');
                 goto 768; end;
        if impme2^.first=nil then begin
                 printimpme(tofile,impme1^.first,impme1^.second);
                 twriteeoln(tofile);
                 twrite(tofile, 'in G1, but this is not the case in G2.|');
                 goto 768; end;
        if same(impme1^.first,impme2^.first) 
            then begin
                 getfirstinpair(impme1^.second,impme2^.second);
                 printimpme(tofile,impme1^.first,impme1^.second);
                 twriteeoln(tofile);
                 if (impme1^.second=impme2^.second) then 
                            twrite(tofile,'in G2, but this is not the case in G1.|')
                       else twrite(tofile,'in G1, but this is not the case in G2.|'); 
                 goto 768;end
            else begin
                 getfirstelem(impme1^.first,impme2^.first);
                 if (impme1^.first=impme2^.first) then 
                       begin
                       printimpme(tofile,impme2^.first,impme2^.second);
                       twriteeoln(tofile);                 
                       twrite(tofile,'in G2, but this is not the case in G1.|');
                       goto 768 end
                  else begin
                       printimpme(tofile,impme1^.first,impme1^.second);
                       twriteeoln(tofile);
                       twrite(tofile,'in G1, but this is not the case in G2.|'); 
                 goto 768;end;
                 end;
            end
       else begin
        twrite(tofile,'Same ancestor relations between imperfect non-conductors,|');
        twrite(output,'Same ancestor relations between imperfect non-conductors,|');
        twriteeoln(tofile);
        twriteeoln(output);
        twrite(tofile,'and pairs of m.e. conductors between the same endpoints, (if any)|');   
        twrite(output,'and pairs of m.e. conductors between the same endpoints, (if any)|');   
        twriteeoln(tofile);
        twriteeoln(output); end;
   
twrite(tofile,'G1 and G2 are equivalent|');
twriteeoln(tofile);
twrite(output,'G1 and G2 are equivalent|');
768: twriteeoln(output);
end;

procedure printgrclass(var tofile:text;
                  var graph:adjtype;
                  var vgraph:adjtype;
                  var impnonconlist:elementptr;
                  var pernonconlist:elementptr;
                  var conlist:elementptr;
                  var pmeconlist:elementptr;
                  var mepairlist:elempairptr;
                  var impimplist:elempairptr;
                  var impmelist:elemelempairptr);

begin;
   twrite(tofile,'1.  Original graph (real edges):|');
   twriteeoln(tofile);
   writegraph(graph,tofile,false,false);
   showgraph(graph,tofile,false);
   twriteeoln(tofile);
   twrite(tofile,'2.  Real and Virtual Adjacencies:|');
   twriteeoln(tofile);
   writegraph(vgraph,tofile,false,false); 
   twriteeoln(tofile);
   showgraph(vgraph,tofile,false);
   twriteeoln(tofile);
 
   printclassif(tofile,conlist,pmeconlist,impnonconlist,pernonconlist);
   printmecons(tofile,mepairlist);
   twriteeoln(tofile);
   
   printimpimplist(tofile,impimplist);
   twriteeoln(tofile);


   printimpmelist(tofile,impmelist);
   twriteeoln(tofile);
   twriteeoln(tofile);
end;

procedure checkequiv;
var graph1,graph2,agraph1,agraph2,vgraph1,vgraph2: adjtype;
{graph1,graph2 contain the underlying 'real' graph                 }
{agraph1,agraph2 contain the 'ancestral relations in the real graph}
{vgraph1,vgraph2 contain the 'virtual and real' adjacencies        }
var conlist1,conlist2,impnonconlist1,impnonconlist2: elementptr;
var pmeconlist1,pmeconlist2,pernonconlist1,pernonconlist2: elementptr;
var mepairlist1,mepairlist2,impimplist1,impimplist2: elempairptr;
var impmelist1,impmelist2:elemelempairptr;

begin
   tnewadj(graph1,fatal);
   tnewadj(graph2,fatal);
   tnewadj(vgraph1,fatal);
   tnewadj(vgraph2,fatal);
   getknowledge;

   copyadjtograph(graph1);
   getknowledge;
   copyadjtograph(graph2);
   makeequivclass(graph1,vgraph1,impnonconlist1,pernonconlist1,conlist1,
                     pmeconlist1, mepairlist1,impimplist1,impmelist1);  
   makeequivclass(graph2,vgraph2,impnonconlist2,pernonconlist2,conlist2,
                     pmeconlist2, mepairlist2,impimplist2,impmelist2);
   compare(ofilvar,vgraph1,vgraph2,
                  impnonconlist1,impnonconlist2,
                  pernonconlist1,pernonconlist2,
                  conlist1,conlist2,
                  mepairlist1,mepairlist2,
                  impimplist1,impimplist2,
                  impmelist1,impmelist2);  
   twriteeoln(ofilvar);
   twriteeoln(ofilvar);
   twrite(ofilvar,'****Classification of G1****|');
   twriteeoln(ofilvar);
   twriteeoln(ofilvar);
   twriteeoln(ofilvar);
   printgrclass(ofilvar,graph1,vgraph1,impnonconlist1,pernonconlist1,conlist1,
                     pmeconlist1, mepairlist1,impimplist1,impmelist1);
   twriteeoln(ofilvar);
   twriteeoln(ofilvar);
   twrite(ofilvar,'****Classification of G2****|');
   twriteeoln(ofilvar);
   twriteeoln(ofilvar);
   twriteeoln(ofilvar);
    printgrclass(ofilvar,graph2,vgraph2,impnonconlist2,pernonconlist2,conlist2,
                     pmeconlist2, mepairlist2,impimplist2,impmelist2);
 

   dispose(impmelist1);dispose(impmelist2);
   dispose(impimplist1);dispose(impimplist2);
   dispose(mepairlist1);dispose(mepairlist2);
   dispose(conlist1); dispose(conlist2);
   dispose(pmeconlist1);dispose(pmeconlist2);
   dispose(impnonconlist1);dispose(impnonconlist2);
   dispose(pernonconlist1);dispose(pernonconlist2);
   dispose(graph1);
   dispose(graph2);
   dispose(vgraph1);
   dispose(vgraph2);
end;








