(*-----------#pc-version-----------------------------------------------------*)
(*implementation*)
(*-----------#pc-version-----------------------------------------------------*)

(*-----------#unix-version---------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "io.h"
#include "sys.h"
#include "util.h"
#include "rsutil.h"
#include "cspecial.h"
#include "scal.h"
#include "calctet.h"
#include "undir.h"
(*-----------#unix-version---------------------------------------------------*)

(* fixed by Peter - October 10, 1991 *)
(* undir.p *)

(*  Procedures for calculating undirected paths *)

  (* These functions and procedures are specialized versions of procedures in calctet and
     calctet1 for calculating undirected paths. *)


(************************************)
(* Calculating the undirected paths *)
     
procedure uddisposepathslist(var paths: udpathsptr; depth: integer);
(* this procedure disposes of a list of paths. *)
label 999;
var temp: udpathsptr;

begin (* disposepathslist *)
   if paths <> nil then
	 begin
	    temp := paths^.next;
            while (paths <> nil) do
	     if ((depth <= paths^.level) or (paths^.level = 0)) then
	       begin
		  dispose(paths);
		  paths := temp;
		  if paths <> nil then temp := paths^.next;
	       end
             else goto 999;
	 end; (* if *)
999: end; (* disposepathlist *)



function udplus(path1,path2: udpathsptr): udpathsptr;
(* this function performs the '+' operation upon two paths and 
   returns the result.  for example, if path1 contains the paths ab and bc
   (representing ab + bc), and path2 contains the paths ef and gh 
   (representing ef + gh), then the function returns the list
   with member ab, cd, ef, and gh (representing ab + cd + ef + gh).  it is a
   destructive function. *)

var curpos: udpathsptr;

begin (* udplus *)
   if aaabort or ((path1 = nil) and (path2 = nil)) then udplus := nil
   else if (path1 = nil) and (path2 <> nil) then
	 udplus := path2
   else if (path1 <> nil) and (path2 = nil) then
	 udplus := path1
      else
	 begin
	    curpos := path1;
	    while curpos^.next <> nil do curpos := curpos^.next;
	    curpos^.next := path2;
	    udplus := path1;
	 end; (* else *)
end; (* udplus *)


function udnointersect(path1,path2: udpathsptr; vert: vertextype): boolean;
begin 
 if (path1 = nil) or (path2 = nil) then udnointersect := true
 else udnointersect := (((path1^.vertices - [vert]) *  
                  (path2^.vertices - [vert])) = []);
end; 


function udconcatpaths(paths1,paths2: udpathsptr; neededges: boolean): udpathsptr;
(* concatenates two paths and returns result.  it is a non-destructive
   function. *)
label 999;
var pathsnew: udpathsptr; temp: parray; 
    i,length1,length2: integer;
begin
   new(pathsnew);
   pathsnew^.level := 0;
   pathsnew^.next := nil;
   pathsnew^.vertices := paths1^.vertices + paths2^.vertices;
   length1 := paths1^.length;
   length2 := paths2^.length;
   if paths1^.patharray[1] = paths2^.patharray[1] then  (* k-i and k-j *)
      begin
        for i := 1 to length1 do
          pathsnew^.patharray[i] := paths1^.patharray[length1 - i + 1];
        for i := 1 to length2 - 1 do
          pathsnew^.patharray[length1+i] := paths2^.patharray[i+1];
      end
   else if paths1^.patharray[1] = paths2^.patharray[length2] then 
              (* k-i,j-k I think this cannot happen *)
      begin
        for i := 1 to length1 do
          pathsnew^.patharray[i] := paths1^.patharray[length1 - i + 1];
        for i := 1 to length2-1 do
          pathsnew^.patharray[length1+i] := paths2^.patharray[length2-i+1];
      end
   else if paths1^.patharray[length1] = paths2^.patharray[1] then  (* i-k-j *)
      begin
        for i := 1 to length1 do
          pathsnew^.patharray[i] := paths1^.patharray[i];
        for i := 1 to length2-1 do
          pathsnew^.patharray[length1+i] := paths2^.patharray[i+1];
      end
   else if paths1^.patharray[length1] = paths2^.patharray[length2] then  (* i-k,j-k *)
      begin
        for i := 1 to length1 do
          pathsnew^.patharray[i] := paths1^.patharray[i];
        for i := 1 to length2-1 do
          pathsnew^.patharray[length1+i] := paths2^.patharray[length2-i];
      end;
   if pathsnew^.patharray[1] > pathsnew^.patharray[length1 + length2 - 1] then
      begin
         for i := 1 to length1 + length2 - 1 do
             temp[i] := pathsnew^.patharray[length1 + length2 - i];
         for i := 1 to length1 + length2 - 1 do
             pathsnew^.patharray[i] := temp[i];
      end;
   pathsnew^.length := paths1^.length + paths2^.length - 1;
   udconcatpaths := pathsnew;
999: end;

  (* this steps down the two paths until they are
     exhausted or there is a difference.  If there is
     a difference in the paths it returns false. *)
function samepath(path1,path2: udpathsptr): boolean;
  var i: vertextype;
      same, finished: boolean;
  begin
    same:= true; finished:= false;
    i:= 0;
    while (same and (not finished)) do
      begin
        i:= i+1;
        same := (path1^.patharray[i] = path2^.patharray[i]);
        finished:= (path1^.patharray[i] = 0) or (path2^.patharray[i] = 0);
      end;
    samepath:= same;
  end;


function udisin(paths1,paths2: udpathsptr; depth: integer): udpathsptr;
(* this function determines whether the list of edges of paths2 occurs
   anywhere in the list headed by paths1 *)
var stop: boolean;
begin
   udisin := nil;
   stop := false;
   while (paths1 <> nil) and (not stop) do
    if (paths1^.level <> 0) and (paths1^.level < depth) then stop := true
    else
      begin
         stop := samepath(paths1,paths2);
	 if (not stop) then paths1 := paths1^.next
         else udisin := paths1;
      end;
end;

function udtimesc(paths1,paths2: udpathsptr; depth: integer;
                  vert: vertextype; neededges: boolean): udpathsptr;
(* this function takes the product of two paths.  it allows only one
   occurence of a given path in the final list.  for example, 
   (ab + cb) * (cd + ad) would return (abcd + abad + cbcd).  it does not
   allow any circular paths to be formed. *)

var temp,curpos1,curpos2,curposnew,newhead: udpathsptr;

begin (* concat *)
   newhead := nil;
   if aaabort or (paths1 = nil)  or (paths2 = nil) then udtimesc := nil
   else
        begin
	    curpos1 := paths1;
	    curposnew := nil;
	    while curpos1 <> nil do
	       begin
		  curpos2 := paths2;
		  while curpos2 <> nil do
		     begin
                        if udnointersect(curpos1,curpos2,vert) then
		      (* if no cycle *)
			      begin
			       temp := udconcatpaths(curpos1,curpos2,neededges);
			       if (udisin(newhead,temp,depth) <> nil) then
			   (* if already in list don't add it *)
				 begin
				  uddisposepathslist(temp,0);
				  temp := nil;
				 end;
			       if temp <>  nil then
				 begin
				  temp^.level := depth;
				  if curposnew = nil then
				    begin
				     curposnew := temp;
				     newhead := curposnew;
				    end
				   else
				    begin
				     curposnew^.next := temp;
				     curposnew := curposnew^.next;
				    end;
				 end (* if *)
			      end; (* if *)
			curpos2 := curpos2^.next;
		     end; (* while *)
		  curpos1 := curpos1^.next;
	       end; (* while *)
	    if curposnew <> nil then
		  begin
		     curposnew^.next := nil;
		     udtimesc := newhead;
		  end (* if *)
	       else udtimesc := nil;
	 end; (* else *)
end; (* concat *)
    
procedure findheadtail(i,j: vertextype; var head,tail: vertextype);
begin
   if i <= j then
      begin
         tail := i;
         head := j;
      end
   else
      begin
         tail := j;
         head := i;
      end;
end;

procedure udcalcpath(*var adjacency: adjtype; var udpathmatrix: udpathstype; 
                     level: integer*); 
(* this procedure calculates the undirected paths between every pair of variables i and j
   and stores the result in udpathmatrix[i,j].  it uses a transitive closure
   algorithm. *)

label 999;
var i,j,k,head,tail,head1,tail1,head2,tail2: vertextype;

begin (* udcalcpath *)
 for i:=1 to max do udpathmatrix^[i,i] := nil;  
 for i:=1 to max-1 do
  begin
      for j:=i+1 to max do
        begin
          if i <> j then
	     begin
              checkinterrupt;
	      if aaabort then goto 999;
              findheadtail(i,j,head,tail);
              udpathmatrix^[i,j] := nil;
	      if (adjacency^[i,j] <> notthere) or
                 (adjacency^[j,i] <> notthere) then
	        begin
	          new(udpathmatrix^[tail,head]);
                  udpathmatrix^[tail,head]^.length := 2;
	          udpathmatrix^[tail,head]^.next := nil;
		  with udpathmatrix^[tail,head]^ do
		    begin
		      vertices := [tail,head];
                      patharray[1]:= tail;
                      patharray[2]:= head;
      		      next := nil;
		    end; (* with *)
	      end (* if *)
	   else udpathmatrix^[i,j] := nil;
        end; (* if *)
     end; (* for *)
   end; (* for *)
   for k:=1 to max do
      for i:=1 to max-1 do
	 for j:=i+1 to max do
	    begin
	       if aaabort then goto 999;
               findheadtail(i,k,head1,tail1);
               findheadtail(j,k,head2,tail2);
	       udpathmatrix^[i,j] := udplus(udpathmatrix^[i,j],
                                           udtimesc(udpathmatrix^[tail1,head1],
                                                    udpathmatrix^[tail2,head2],level,k,true));
	       if udpathmatrix^[i,j] <> nil then
		     udpathmatrix^[i,j]^.level := 1;
	    end;
999: end; (* UdCalcpath *)

(* End of Calculating the undirected paths *)
(*******************************************)
(*-----------#pc-version-----------------------------------------------------*)
(*end.*)
(*-----------#pc-version-----------------------------------------------------*)