(*-----------#pc-version-----------------------------------------------------*)
(*
implementation
uses util,rsutil,io,mon,sys;
*)
(*-----------#pc-version-----------------------------------------------------*)

(*-----------#unix-version---------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "mon.h"
#include "up.h"
#include "calctet.h"
#include "io.h"
#include "sys.h"
#include "util.h"
#include "eqs.h"
#include "cspecial.h"
#include "rsutil.h"
#include "testbuildm.h"
#include "cross.h"
(*-----------#unix-version---------------------------------------------------*)

(******************************************************************)
(*  History and updates                                           *)
(*                                                                *)
(*                                                                *)
(*                                                                *)
(*  6/17/91   - completed first version                           *)
(*                                                                *)
(******************************************************************)


(******************************************************************)
(*  History and updates                                           *)
(*                                                                *)
(*  10/11/91  - changed printing of update output                 *)
(*  6/26/91   - added analytic code                               *)
(*  6/17/91   - completed first version                           *)
(*                                                                *)
(******************************************************************)


(***********************************************************************)
const 
      numofexplanations = 32;            {how many top explanations to find}
      testcount = 200;                   {how often to test for convergence}

type

   explanation = record
	instantiation:vertarray;
	probability:real;
	end;
   varray = array[vertextype] of vertextype;

var
   target: vertextype;

   topexplanations : array[1..numofexplanations] of explanation;
   lstc : inttallycard;            {tallycard for counts in logic sampling}
   samplecount : integer;
   hitcount,numz : integer;


   testnumber : integer;
   condprob:real;
   marg1,marg2: array[0..maxcategories] of real;
   marg: array[0..maxcategories,0..maxcategories] of real;

(*************************************************************************)

(***************************************************************************)
(* i/o section  There is a separate output section below                   *)
(* functions in this section should probably be in io.p                    *)
(***************************************************************************)

(* replaces all occurances of c with a space character *)
{procedure replacecharwithspace(c:char);
   var
	i:integer;
   begin
	for i := 1 to buflength do
		if buf[i]=c
		then buf[i] := ' ';
   end;
}

(************************************************************************)
(* utilities section                                                    *)
(************************************************************************)

(* gets the probability of the effect having value given the cutoffs *)
(* returns the probability of the effect having taken on the value for
   this set of cutoffs.  the set of cutoffs is determined in getjointprob
   based on the current sample *)

function getprobability(effect:integer; cutoffs:probarray; value:real):real;
   var
	category:integer;
	probval:real;
   begin
	category:=getnumcat(effect,value,cats);
        if cutoffs[1] <> -1 then
             if (category = cats[effect].num) then
                probval:= 1 - cutoffs[category - 1]
	     else probval:= cutoffs[category] - cutoffs[category - 1]
        else probval := 1/cats[effect].num;
	getprobability:=probval;
   end;

(* get the joint probability of the sample based on the 
   parameterization in CATS *)

function getjointprob(sample:vertarray):real;
   label 999; 
   var
	probval:real;
	effect:integer;
	pptr: probptr;
	cutoffs: probarray;

   begin
	probval:=1.0;
	for effect :=1 to origmax do  
           if effect in graphvertices then
	   begin
		pptr:= cats[effect].condprob;
		findcondprob(pptr,effect,cutoffs,cats,sample);
		probval := probval * getprobability(effect,cutoffs,sample[effect]);
                if probval = 0 then goto 999;
	   end;
999:	getjointprob:= probval;
   end;

procedure printsampleandprob(sample:vertarray;probval:real);
   var
	i:integer;
   begin
	twriteeoln(output);
	for i := 1 to origmax do
	begin
	   tpwritereal(output,sample[i],1,0,false);twrite(output,' |');
	end;
	twrite(output,'   |');
	tpwritereal(output,probval,6,5,false);
	twriteeoln(output);
	twriteeoln(output);
   end;

(************************************************************************)
(* END of utilities section                                             *)
(************************************************************************)

(* evidence section                                                     *)
(*                                                                      *)
(* functions for obtaining, checking, and manipulating evidence         *)
(*                                                                      *)
(* evidencelist:observationlist;                                        *)
(*    evidencelist is an ORDERED LINKED LIST of observations            *)
(*    the order is based on the ordering of vernames which is an        *)
(*    array of vertexnames                                              *)
(* evidenceset:vertexset;                                               *)
(*    evidenceset is the set of current evidence                        *)
(************************************************************************)

function valueof(number:vertextype;elist:observationlist):real;
   begin
	elist := elist^.next;  {skip dummy observation}
	while (elist <> nil) and (elist^.number<>number) do
		elist:=elist^.next;
	if elist = nil then valueof:=0
	else valueof:= elist^.value;
   end;

function thereISevidence:boolean;
   begin
	thereISevidence:=(not (evidencelist^.next = nil));
   end;

(*check to see if two samples agree on on values
  returns true if samples agree                     *)

function matchsamples(sample1,sample2:vertarray):boolean;
   var
	i : integer;
	mismatch:boolean;
   begin
	mismatch := false;
	i:=1;
	while ((i<=origmax) and (not mismatch)) do
           if (sample1[i]=sample2[i])
           then i := i+1
	   else mismatch := true; 
	matchsamples := (not mismatch);
   end;

(* checks to see if the sample corresponds with the evidence
   returns TRUE if it does match *)
function matchsampletoevidence(elist:observationlist; 
                               sample:vertarray):boolean;
   var
	mismatch:boolean;
   begin
	mismatch := false;
	elist := elist^.next;  {skip dummy observation}
	while ((elist<>nil) and (not mismatch)) do
	   begin
		if (elist^.value = sample[elist^.number])
		then elist:=elist^.next
		else mismatch := true;
	   end;
	matchsampletoevidence := (not mismatch);
   end;


(* changes a sample (vertex array) to correspond with the evidence *)
procedure changesamplewithevidence(elist:observationlist;
                                   var sample:vertarray);
   begin
	elist := elist^.next;  {skip dummy observation}
	while (elist<>nil) do
	   begin
		sample[elist^.number] := elist^.value;
		elist:=elist^.next;
	   end;
   end;

procedure insertevidence(var evidencelist:observationlist;
			name: vertexnametype; number:vertextype; value:real;
			var inserted:boolean);
   var	tempptr:observationlist;

   begin
	tempptr := evidencelist^.next;
	new(evidencelist^.next);
	evidencelist := evidencelist^.next;
	evidencelist^.next := tempptr;
	evidencelist^.number := number;
	evidencelist^.name := name;
	evidencelist^.value := value;
	inserted := true;
   end;

(* add a piece of evidence into an orderedlinked list
   if evidence for a vertex exists then it changes the value
   if evidence for a vertex does not exist then it creates an observation
      and places it in the list *)
procedure addevidence(*evidencelist: observationlist;
                      var evidenceset:vertexset;
                      name: vertexnametype; number:vertextype; value:real*);
   var	
	inserted: boolean;
   begin
	evidenceset:=evidenceset + [number];
	inserted := false;
	while (not inserted) do
		{ if we hit the end of the list or find a 
                  variable who's number is greater
		  then we insert the new variable}
	   if (evidencelist^.next = nil)
	   then
		insertevidence( evidencelist,name, number, value,inserted)
	   else
		if (evidencelist^.next^.number > number)
		then insertevidence(evidencelist,name,number,value,inserted)
		else

		if (evidencelist^.next^.number = number) then
		   begin
			evidencelist^.next^.value := value;
			{ value allready exists for variable }
			inserted := true;
		   end
		else	
		   begin
			{move pointer down list}
			evidencelist := evidencelist^.next;
		   end;
   end;


(* checks to see if the value is a proper category for the
   vertex in cats[number]  *)
function badcategory(number:vertextype;value:real):boolean;
   var
	i:integer;
	found:boolean;
   begin
     if number <> 0 then
      begin
	found:=false;
	i:=0;
        while ((not found) and (i<cats[number].num)) do
	   begin
		i:=i+1;		{array starts at 1}
		if (cats[number].value[i]=value) then found:=true;
	   end;
	badcategory:= (not found);
      end;
   end;


(* this procedure querries the user for observations/evidence
   and checks if the values obtained are acceptable values for the variables
   depends on global: VERNAMES *)
procedure getevidence(evidencelist: observationlist;
                      var evidenceset:vertexset);   
   label 999;
   var 
     curname: vertexnametype;
     curnumber: vertextype;
     curvalue: real;
     wasinserted,emptyline: boolean;

   begin
     emptyline := false;
     if (currentin = inp) and (max > 0) then
        begin
           twriteln(output,'Example of input: |');
           writename(1,output);
           twrite(output,' |');
           tpwritereal(output,cats[1].value[1],5,0,false);
           twriteeoln(output);
        end;
     repeat
        errors := [];
	if (currentin = inp) then
           begin
             twrite(output,'Evidence: |');
             readttyline(buf);
           end
        else readfileline(knowfile,buf);
        if myeoln then emptyline := true
        else
           begin
		readwhitespace;
		readname(curname, ignoreall);
                if not (ignoreall in errors) then
                  begin
    		   curnumber := listposition(curname,false,vernames,
                                             wasinserted);
                   if curnumber = 0 then goto 999;
                   errors := [];
		   curvalue:= readnum(false,ignoreall);
                   if ignoreall in errors then goto 999;
		   {	  check if valid value    }
                   if curnumber in evidenceset then 
		       myerror('Evidence already entered for this variable. Overwriting previous value.|',
		       	       warning,false);
		   if (not badcategory(curnumber,curvalue)) and 
                      (not (ignoreall in errors)) then
		      addevidence(evidencelist,evidenceset,curname,
                                  curnumber,curvalue)
		   else	
		   begin
			twrite(output,'Warning: |');
			twriteint(output,tround(curvalue));
			twrite(output, ' is not a proper category for |');
			writevername(curname,output);twriteeoln(output);
			twrite(output,'Not adding |');
			writevername(curname,output);
			twrite(output, ' = |');
			twriteint(output, tround(curvalue));
			twrite(output,' to evidence list|');
			twriteeoln(output);
		    end;
                   end;
                checkend;
999:	   end;
     until emptyline;
   end;	

(* removes all evidence from the current evidence list and returns
   all of the allocated space back to the heap *)
procedure purgeevidence(elist:observationlist;var eset:vertexset);
   var	tempptr : observationlist;

   begin
	elist := elist^.next;   {skip dummy first observation}
	while (elist <> nil) do
	   begin
		tempptr := elist;
		elist := elist^.next;
		dispose(tempptr);
	   end;
	evidencelist^.next := nil;
	eset:=[];
   end;

(* given the name of a vertex this procedure will remove the observation
   that corresponds to that vertex and returns the space to the heap
	depends on global: VERNAMES *)
procedure removepieceofevidence(elist:observationlist;
                         var evidenceset:vertexset; name:vertexnametype);
   label 999;
   var	tempptr : observationlist;
	number : vertextype;
        wasinserted: boolean;

   begin
	number := listposition(name,false,vernames,wasinserted);
        if number = 0 then goto 999;
        if (not (number in evidenceset)) then 
           myerror('Variable not in evidence set.|',
          	    ignoreall,false)           
        else
          begin
            evidenceset:=evidenceset-[number];
	    while (elist^.next <> nil) do
              begin
		if (elist^.next^.number = number) 
                        { we found evidence to remove}
		then
		   begin
			tempptr := elist^.next;
			elist^.next := tempptr^.next;
			dispose(tempptr);
		   end
		else
		   elist := elist^.next;
  	      end; 
     end;
999:   end;

(* this procedure querries the user for which pieces of evidence to remove and
   then removes those pieces of evidence *)
procedure removeevidence(elist:observationlist;var eset:vertexset);
   var  emptyline: boolean;
	name:vertexnametype;
begin
  if elist^.next = nil then
    twriteln(output,'No evidence entered.|')
  else
   begin
     emptyline := false;
     if (currentin = inp) and (max > 0) then 
        begin
           twriteln(output,'Example of input: |');
           writename(elist^.next^.number,output);
           twriteeoln(output);
        end;
     repeat
     errors := [];
     if (currentin = inp) then
       begin
          twrite(output,'Variable: |');
          readttyline(buf) 
       end
     else readfileline(knowfile,buf);
     if myeoln then emptyline := true 
     else
	   begin
		readname(name, ignoreall);
		if not (ignoreall in errors) then 
                   removepieceofevidence(elist,eset,name);
	   end;
     checkend;
    until emptyline;
   end;
end;
(* initializes the evidence section by creating a dummy head variable *)
procedure initevidence(var evidencelist:observationlist);

   begin
	new(evidencelist);   {make dummy node}
	evidencelist^.next := nil;
	evidenceset:=[];
   end;
 

(***************************************************************************)
(* END of evidence section                                                 *)
(***************************************************************************)

(***************************************************************************)
(* Most probable explanation section                                       *)
(*                                                                         *)
(*  topexplanations : array[1..numofexplanations] of explanation;          *)
(*      topexplanations is an array of explanations (samples and probs)    *)
(*      ordered by the probability of the samples.                         *)
(*                                                                         *)
(*  by calculating the joint probability of the samples generated in       *)
(*  the process of logic sampling we will (with high probability) find     *)
(*  the instantiation (sample) that best fits the evidence.  The sample    *)
(*  with the highest joint probability                                     *)
(*                                                                         *)
(*  if updating analytically then the topexplanations array will contain   *)
(*  the most probable instantiations.                                      *)
(***************************************************************************)


(* given a sample and prob this procedure will insert the sample into the
   appropriate position of the TOPEXPLANATIONS array while adjusting the
   other entries in the array properly.  An entry can only appear once in
   the array *)
procedure insertexplanation(sample:vertarray;probval:real);
var i,j:integer;
    inserted:boolean;

   begin
	i:=numofexplanations;
	inserted := false;
	if (probval<topexplanations[i].probability)
	then inserted := true;

	{check if explanation is already inserted}
	if not inserted then
	   begin
		for j:=1 to numofexplanations do
		   if (topexplanations[j].probability = probval) and
			 (matchsamples(topexplanations[j].instantiation,sample))
		   then inserted := true;
	   end;

	while (not inserted) do
	   begin
		i := i-1;
		if ((i = 0) or (probval<=topexplanations[i].probability))
		then
		   begin
			for j:=1 to maxvertex do
				topexplanations[i+1].instantiation[j] := sample[j];
			topexplanations[i+1].probability := probval;
			inserted := true;
		   end
		else
		   begin
			for j:=1 to maxvertex do
				topexplanations[i+1].instantiation[j] := 
					topexplanations[i].instantiation[j];
			topexplanations[i+1].probability := topexplanations[i].probability;
		   end;
	   end;
   end;


(* erases all of the fields in the topexplanations array *)
procedure erasetopexplanations;
var i,j:integer;
   begin
	for i:=1 to numofexplanations do
	   begin
		for j:=1 to origmax do
		   topexplanations[i].instantiation[j] := 0;
		topexplanations[i].probability := 0;
	   end;
	(*twriteeoln(output);*)
   end;

(***************************************************************************)
(* END of Most probable explanation section                                *)
(***************************************************************************)



(***************************************************************************)
(* CONVERGENCE SECTION                                                     *)
(***************************************************************************)

procedure fillcard(countcard:inttallycard;hitcount:integer;
                   var freqcard:realtallycard);
   var
	vertex,category:integer;
   begin
	for vertex:=1 to origmax do
	   for category:= 1 to cats[vertex].num do
		freqcard[vertex,category]:=countcard[vertex,category]/hitcount;
   end;

procedure fillrealcard(countcard:realtallycard;hitcount:real;
                       var freqcard:realtallycard);
   var
	vertex,category:integer;
   begin
        if hitcount <> 0 then
	for vertex:=1 to origmax do
	   for category:= 1 to cats[vertex].num do
             freqcard[vertex,category]:=countcard[vertex,category]/hitcount;
   end;

procedure copycard(fromcard:realtallycard; var tocard:realtallycard);
   var
	vertex,category:integer;
   begin
	for vertex:=1 to origmax do
	   for category:= 1 to cats[vertex].num do
		tocard[vertex,category]:=fromcard[vertex,category];
   end;

procedure measurediff(card1,card2:realtallycard;var maxdiff,totaldiff:real);
   var
	diff:real;
	vertex,category:integer;
   begin
	totaldiff:= 0;
	maxdiff:= 0;
	for vertex:=1 to origmax do
	   for category:= 1 to cats[vertex].num do
	   begin
	   	if card1[vertex,category]<>0 then
		diff:=abs(card1[vertex,category]-
                          card2[vertex,category])/card1[vertex,category];
		totaldiff:= totaldiff+diff;
		if diff>maxdiff then maxdiff:=diff;
	   end;
   end;


function measurederror(card1,card2:realtallycard):real;
   var
	error,totalerror:real;
	vertex,category:integer;
	norm:integer;
   begin
	totalerror:=0;
	norm:=0;
	for vertex:=1 to origmax do
	   for category:= 1 to cats[vertex].num do
	   begin
		norm:=norm+1;
		error:=abs(card1[vertex,category]-card2[vertex,category]);
		totalerror:= totalerror+error;
	   end;
	measurederror:= totalerror/norm;
   end;

function shouldIquit:boolean;
   var
	maxdiff,totaldiff:real;
   begin
	measurediff(freqcard,ofreqcard,maxdiff,totaldiff);
  {	shouldIquit:=(samplecount>500);}

	if maxdiff<0.0001
	then shouldIquit:=true
	else shouldIquit:=false;

 {	if totaldiff<(0.02 * origmax)
	then shouldIquit:=true
	else shouldIquit:=false;
 }
   end;

procedure testconvergence(var outfile:text);
   var
	error1,error2:real;
   begin
	fillcard(lstc,hitcount,freqcard);
	fillrealcard(lwtc,likelihoodcount,lwfreqcard);
	error1:=measurederror(exactfreqcard,freqcard);
	error2:=measurederror(exactfreqcard,lwfreqcard);

	twriteint(outfile,samplecount);twrite(outfile,'  |');
	twriteint(outfile,hitcount);twrite(outfile,'  |');
	tpwritereal(outfile,error1,10,8,false);twrite(outfile, '  |');
	tpwritereal(outfile, error2,10,8,false);twriteeoln(outfile);
   end;

(***************************************************************************)
(* END of convergence section                                              *)
(***************************************************************************)

(***************************************************************************)
(* tally card section                                                      *)
(*   inttallycard = array[1..maxvertex, 1..maxcategories] of integer;      *)
(*   realtallycard = array[1..maxvertex, 1..maxcategories] of real;        *)
(*                                                                         *)
(*   tallycards are used to keep track of values for the different         *)
(*   categories of each variable.  the inttallycard is used in the logic   *)
(*   sampling and the realtallycard is used in the convergence section     *)
(*   and the likelihood weighting section                                  *)
(*   exactfreqcard, if used, will contain the exact relative frequencies   *)
(*   freqcard, ofreqcard are used in the convergence section above         *)
(***************************************************************************)

(* erase integer tallycard *)
procedure eraseinttallycard(var card:inttallycard);
   var
	vertex, category:integer;

   begin
	for vertex:=1 to maxvertex do
	   for category := 1 to maxcategories do
		card[vertex,category]:=0;
   end;

(* erase real tallycard *)
procedure eraserealtallycard(var card:realtallycard);
   var
	vertex, category:integer;

   begin
	for vertex:=1 to maxvertex do
	   for category := 1 to maxcategories do
		card[vertex,category]:=0;
   end;


(* takes a sample and increments the tallycard by one for the value taken on by
   a variable in the sample.*)
procedure inttally(sample:vertarray;var card:inttallycard);
   var
	effect, category:integer;
   begin

	for effect :=1 to origmax do 
	   begin
		category:=getnumcat(effect,sample[effect],cats);
		card[effect,category]:=card[effect,category] + 1;
	   end;
   end;

(* takes a sample and increments the tallycard by weight for the value taken on by
   a variable in the sample.*)
procedure realtally(sample:vertarray;var card:realtallycard; weight:real);
   var
	effect,category:integer;
   begin

	for effect :=1 to origmax do 
	   begin
		category:=getnumcat(effect,sample[effect],cats);
		card[effect,category]:=card[effect,category] + weight;
	   end;
   end;

procedure normalizereal(var card:realtallycard);
   var
	effect,category:integer;
	sum:real;
   begin
	for effect :=1 to origmax do 
	   begin
		sum:=0;
		for category:=1 to cats[effect].num do
		   sum:=sum+card[effect,category];
		if sum <> 0 then
		  for category:=1 to cats[effect].num do
		     card[effect,category]:=card[effect,category]/sum;
	   end;
   end;



(***************************************************************************)
(* END tally card section                                                  *)
(***************************************************************************)

(***************************************************************************)
(* logic sampling section                                                  *)
(***************************************************************************)


(* gets a sample; an assignment for all of the variables in the graph

   Depends on the globals varorder and cats
   varorder is set in formandprintequations (monte.p)
   cats is set in getbayesnet ? (monte.p) *)

procedure getlogicsample(var sample:vertarray);
   var
	frontptr: eqorder; 
	effect: integer;

   begin
	frontptr:= varorder;
	while (frontptr^.next <> nil) do
	   begin
		effect := frontptr^.varbl;
		getdiscreteval(effect,sample,cats);  { get value for effect }
		frontptr:=frontptr^.next;
	   end;
   end;


procedure DOlogicsampling;
   var
	converged:boolean;
	sample:vertarray;
	probval:real;
   begin
	while (not converged) do
	   begin
		getlogicsample(sample);
		samplecount := samplecount + 1;
		if (matchsampletoevidence(evidencelist,sample))
		then
		   begin    {a hit for logic sampling }
			hitcount := hitcount + 1;
			inttally(sample,lstc);
		   end
		else
		   changesamplewithevidence(evidencelist,sample);
		probval:=getjointprob(sample);

		insertexplanation(sample,probval);			{ DO Most probable explanation }

		if ((hitcount mod 30)=0) 
		then
		   begin
			twrite(output,'.|');
			flush(output);
		   end;   

		if ((hitcount mod testcount)=0) 
		then
		   begin
			copycard(freqcard,ofreqcard);
			fillcard(lstc,hitcount,freqcard);
			converged:=shouldIquit;
		   end;
	   end;
   end;

(***************************************************************************)
(* END of logic sampling section                                           *)
(***************************************************************************)

(***************************************************************************)
(* likilihood weighting section                                            *)
(***************************************************************************)


procedure getlikelihoodsample(var sample:vertarray;eset:vertexset;elist:observationlist);
   var
	frontptr: eqorder; 
	effect: vertextype;

   begin
	frontptr:= varorder;
	while (frontptr^.next <> nil) do
	   begin
		effect := frontptr^.varbl;

		if effect in eset
		then sample[effect]:=valueof(effect,elist)
		else getdiscreteval(effect,sample,cats);  { get value for effect }

		frontptr:=frontptr^.next;
	   end;
   end;



(* get the score for this sample *)
function getlikelihoodscore(* sample:vertarray;evidenceset:vertexset):real;*);
var
	probval:real;
	effect:vertextype;
	pptr: probptr;
	cutoffs: probarray;

   begin
	probval:=1.0;
	for effect :=1 to origmax do  
	   begin
		if effect in evidenceset
		then
		   begin
			pptr:= cats[effect].condprob;
			findcondprob(pptr,effect,cutoffs,cats,sample);
			probval := probval * getprobability(effect,cutoffs,sample[effect]);
		   end;
	   end;
	getlikelihoodscore:=probval;
   end;


procedure DOlikelihoodweighting;
   var
	weight:real;
	converged:boolean;
	sample:vertarray;
        probval:real;
   begin
	twriteln(output,'Using likelihood weighting method.|');
	converged:=false;
	while (not converged) do
	   begin
		getlikelihoodsample(sample,evidenceset,evidencelist);	{ DO likelihood update }
		weight:=getlikelihoodscore(sample,evidenceset);
		likelihoodcount:=likelihoodcount + weight;
		realtally(sample,lwtc,weight);
		probval:=getjointprob(sample);
		insertexplanation(sample,probval);			{ DO Most probable explanation }

		if ((samplecount mod 30)=0) 
		then
		   begin
			twrite(output,'.|');
			flush(output);
		   end;   

		if ((samplecount mod testcount)=0) 
		then 
		   begin
			copycard(freqcard,ofreqcard);
			fillrealcard(lwtc,likelihoodcount,freqcard);
			converged:=shouldIquit;
		   end;
	   end;
	   twriteeoln(output);
   end;



(***************************************************************************)
(* END OF likilihood weighting section                                     *)
(***************************************************************************)

(*****************************************************************************)
(* Analytic solution section                                                 *)
(*****************************************************************************)

function getstatespace:integer;
   label 999;
   var
	i:vertextype;
	space:integer;
   begin
	space:=1;
	for i:=1 to origmax do
	   begin
		if ((not (i in evidenceset)) and (i in graphvertices)) then space := space * cats[i].num;
		if space>maxstatespace then goto 999;
	   end;
999:	getstatespace:=space;
   end;

procedure decode(code,size:integer;var sample:vertarray; numUnknown:integer; unknownarray:varray; 
                 var cats: categoryarray);
 
   var num:integer;
	vertex:vertextype;
	i:integer;
   begin

	{put in known values}
	for vertex:=1 to origmax do
	   if vertex in evidenceset
	   then sample[vertex]:=valueof(vertex,evidencelist);

	{get unknown values}

	for i:= 1 to numUnknown do
	   begin
		vertex:=unknownarray[i];
		size:= tround(size / cats[vertex].num);
		num:=(code div size);
		sample[vertex]:=cats[vertex].value[num+1];
		code:=code-(size*num);
	   end;
   end;


procedure marginalize(size:integer;unknownarray:varray);
   var
	i,j,k,l:integer;
	probval:real;
	vertex:vertextype;
	above,inarow,below,ncat:integer;

   begin
	for i:= 1 to origmax do
           if i in graphvertices then
	   begin
		ncat:=cats[i].num;
		if i in evidenceset
		then
		   for j:= 1 to cats[i].num do
		      begin
			if (j = getnumcat(i,valueof(i,evidencelist),cats))
			then exactfreqcard[i,j]:=1
			else exactfreqcard[i,j]:=0;
		      end
		else
		   begin
			k:=0;
			inarow:=size;
			repeat
			   k:=k+1;
			   vertex:=unknownarray[k];
			   inarow:=tround(inarow/cats[vertex].num);
			until unknownarray[k] = i;

			for j:= 1 to cats[i].num do
			   begin
				probval:=0;
				above:=inarow * (j-1);
				below:=inarow * (ncat - j);
				k:=0;
				while k<size do
				   begin
					k:= k+above;
					for l:= 1 to inarow do
					   probval:=probval + statespace^.space[k+l-1];
					k:=k+inarow;
					k:=k+below;
				   end;
				exactfreqcard[i,j]:=probval;
			   end;
		   end;
	   end;
   end;

procedure DOexact(size:integer);
   var
	i:vertextype;
	j:integer;
	numUnknown:integer;
	unknownarray:varray;
	sample:vertarray;
	probval:real;
   begin
	statespace^.size:=size;
	numUnknown:=sizeofset(graphvertices)-sizeofset(evidenceset);
	j:=0;

	{set up unknownarray}
	for i:= 1 to origmax do
		if ((not (i in evidenceset)) and (i in graphvertices)) then
		   begin
			j:=j+1;
			unknownarray[j]:=i;
		   end;

	{	fill in statespace  }
	condprob:=0;
	for j:= 0 to (size - 1) do
	   begin
		decode(j,size,sample,numUnknown,unknownarray,cats);
		probval:=getjointprob(sample);
		insertexplanation(sample,probval);			{ DO Most probable explanation }
		condprob:=probval+condprob;
		statespace^.space[j]:=probval;
	   end;

	{marginalize and fill in exactfreqcard}
	marginalize(size,unknownarray);
	normalizereal(exactfreqcard);
   end;
  
(*****************************************************************************)
(* END of Analytic solution section                                          *)
(*****************************************************************************)

(***************************************************************************)
(* update control section                                                  *)
(***************************************************************************)

procedure initupdate;
   begin
	hitcount:=0;
	samplecount := 0;
	likelihoodcount := 0;
	eraseinttallycard(lstc);
	eraserealtallycard(lwtc);
	eraserealtallycard(freqcard);
	eraserealtallycard(ofreqcard);
	eraserealtallycard(exactfreqcard);
	erasetopexplanations;
   end;



(* This procedure updates the network and depends on VARORDER and CATS*)
(* The update is done with likelihoodweighting(can change to logic sampling)
   The Most Probable Explanation is done by forcing each sample to match the evidence
       and then accessing the joint probability of that sample.  See get likelihood sample.
*)


procedure updatebayes;

   var
	space:integer;
	starttime,endtime:real;
   begin

{	twriteln(output,'Updating network|');}
	initupdate;
	space:=getstatespace;

	if space>maxstatespace 
	then
	   begin
		DOlikelihoodweighting;
		updatemethod:=likelihoodweighting;
	   end
	else 
	   begin
		DOexact(space);
		updatemethod:=analytic;
	   end;
	endtime:=tgettime;
	{writeln(output,'Total update time: ',endtime-starttime);}
   end;

(***************************************************************************)
(* END of update control section                                           *)
(***************************************************************************)

(****************************************************************************)
(* OUTPUT section                                                           *)
(****************************************************************************)

procedure writeheader(var outfile:text);
   begin
	twriteln(outfile,'{|');
	twriteln(outfile,'Output of updating netowrk|');
	twriteln(outfile,'}|');
   end;

procedure printevidence(elist: observationlist;var outfile:text);

   begin
	twriteeoln(outfile);
	twriteeoln(outfile);
	if thereISevidence
	then twriteln(outfile,'The current list of evidence is:|');
	while (not (elist^.next = nil)) do
	   begin
		{skip the dummy}
		elist:=elist^.next;
		writevername(elist^.name,outfile); twrite(outfile,'   |');
		twriteint(outfile,tround(elist^.value));
		twriteeoln(outfile);
	   end;
	twriteeoln(outfile);
	elist := nil;
   end;

(* write out evidence in one line *)
(* x=1,y=2,z=0 *)
procedure writeevidence(elist: observationlist;var outfile:text);
   var
	done:boolean;
   begin
	done:=false;

	if (elist^.next = nil)
	then twrite(outfile,'T|');

	while (not done) do
	   begin
		elist:=elist^.next;		{skip the dummy}
		done:=(elist^.next = nil);
		writename(elist^.number,outfile);
		twrite(outfile,'=|');
		twriteint(outfile,tround(elist^.value));
		if (not done) then twrite(outfile,', |');
	   end;
	elist := nil;
   end;
(* The test for cprobset, a boolean value that is set based on the method
   being used for updating, is in the code because the  likelihood weighting
   method does not calculate the prob. of the conditioning set so cprobset is
   set to false.  If logic sampling is added as a possible method then
   one should add back in the two commented out section that write out the
   joint probability of the instantiation. *)
procedure printtopexplanations(cprob:real;cprobset:boolean;var outfile:text);
var i,j:integer;
 begin
  if cprob <> 0 then
    begin
    	twriteeoln(outfile);
	twriteln(outfile,'The following assignments are the best explanations|');
	twriteeoln(outfile);
	for j:=1 to origmax do
	   pwritename(j,outfile,6);
	twriteln(outfile,'  prob. for instantiation|');
	twriteeoln(outfile);
	for i:=1 to numofexplanations do
	 begin
	    if (topexplanations[i].probability > 0.0) then
	      begin
	      	for j:=1 to origmax do
		     tpwriteint(outfile,tround(topexplanations[i].instantiation[j]),6,false);
		twrite(outfile,' |');
		tpwritereal(outfile,topexplanations[i].probability,8,4,false);
		twriteeoln(outfile);
	       end;
	   end;
	twriteeoln(outfile);
	end;
   end;

procedure printinttallycard(tc:inttallycard;norm:integer;var outfile:text);
   var
	i,j:integer;
   begin
	for i:=1 to origmax do
	   begin
		for j := 1 to cats[i].num do
		   begin
                        if (j > 1) and (((j-1) mod 3) = 0) then
                               twriteeoln(outfile);
			twrite(outfile,'P(|');
			writename(i,outfile);
			twrite(outfile,'=|');
			twriteint(outfile,tround(cats[i].value[j]));
			if thereISevidence 
			then 
			   begin
				twritechar(outfile,'|');
				writeevidence(evidencelist,outfile);
			   end;
			twrite(outfile,') = |');
			tpwritereal(outfile,tc[i,j]/norm,5,4,false);
			twrite(outfile,'     |');
		   end;
		twriteeoln(outfile);
	   end;
   end;


procedure printrealtallycard(tc:realtallycard;norm:real;var outfile:text);
   var
	i,j:integer;
   begin
	for i:=1 to origmax do
	   begin
		for j := 1 to cats[i].num do
		   begin
                        if (j > 1) and (((j-1) mod 3) = 0) then
                               twriteeoln(outfile);
			twrite(outfile,'P(|');
			writename(i,outfile);
			twrite(outfile,'=|');
			twriteint(outfile,tround(cats[i].value[j]));
			if thereISevidence 
			then 
			   begin
				twritechar(outfile,'|');
				writeevidence(evidencelist,outfile);
			   end;
			twrite(outfile,') = |');
			tpwritereal(outfile,tc[i,j]/norm,5,4,false);
			twrite(outfile,'     |');
		   end;
		twriteeoln(outfile);
	   end;
   end;


procedure printlwtally(var outfile:text);
   begin
	twriteeoln(outfile);
	twriteln(outfile,'Estimated conditional probabilities (likelihood weighting)|');
	twriteeoln(outfile);
	printrealtallycard(lwtc,likelihoodcount,outfile);
	twriteeoln(outfile);
   end;

procedure printlstally(var outfile:text);
   begin
	twriteeoln(outfile);
	twriteeoln(outfile);
	if thereISevidence
	then
	   begin
		twriteln(outfile,'Estimation of the probability of the conditioning set|');
		twriteeoln(outfile);
		twrite(outfile,'P(|');
		writeevidence(evidencelist,outfile);
		twrite(output,') ~= |');
		tpwritereal(output,hitcount/samplecount,7,5,false);
		twriteeoln(output);
	   end
	else
	   begin
		twriteln(outfile,'Null conditioning set|');
		twriteeoln(outfile);
	   end;
	twriteeoln(outfile);
	twriteeoln(outfile);
	twriteln(outfile,'Estimated conditional probabilities (logic sampling)|');
	twriteeoln(outfile);
	printinttallycard(lstc,hitcount,outfile);

	twriteeoln(outfile);
	twriteeoln(outfile);
	twriteln(outfile,'Estimated conditional probabilities (likelihood weighting)|');
	twriteeoln(outfile);
	printrealtallycard(lwtc,likelihoodcount,outfile);
	twriteeoln(outfile);
   end;

procedure writecsetprob(cprob:real;var outfile:text);
begin
	twriteln(outfile,'The probability of the conditioning set|');
	twriteeoln(outfile);
	twrite(outfile,'P(|');
	writeevidence(evidencelist,outfile);
	twrite(outfile,') = |');
	tpwritereal(outfile,cprob,7,4,false);
	twriteeoln(outfile);
end;


procedure printexacttally(var outfile:text);
   begin
	twriteeoln(outfile);
	if thereISevidence
	then writecsetprob(condprob,outfile)
	else
	   begin
		twriteln(outfile,'Null conditioning set|');
		twriteeoln(outfile);
	   end;
	twriteeoln(outfile);
	twriteln(outfile,'Conditional probabilities (marginalizing statespace)|');
	twriteeoln(outfile);
	printrealtallycard(exactfreqcard,1,outfile);
	twriteeoln(outfile);
   end;

procedure printbayesoutput(var outfile:text);
var
	cprob:real; 
	cprobset:boolean; 
begin
   printevidence(evidencelist,outfile);
   cprobset:=false;
   if updatemethod=logicsampling then
    begin
	cprobset:=true;
	cprob:=hitcount/samplecount
    end
   else if updatemethod=analytic then
    begin
	cprobset:=true;
	cprob:=condprob;
    end
   else cprob:=likelihoodcount;
   if cprob <> 0 then
     begin
	printtopexplanations(cprob,cprobset,outfile);
	if updatemethod=analytic then printexacttally(outfile)
	else if updatemethod=likelihoodweighting then printlwtally(outfile)
	else if updatemethod=logicsampling then printlstally(outfile);
      end
   else writecsetprob(0,outfile);
   flush(outfile);
   end;

procedure printbayesoutput1(var outfile:text);
var curvalue,maxvalue: real; curid,i: integer;
begin
  maxvalue := -1;
  for i := 1 to cats[target].num do
   begin
     if updatemethod=analytic then curvalue := exactfreqcard[target,i]
     else if updatemethod=likelihoodweighting then curvalue := lwtc[target,i]/likelihoodcount
     else if updatemethod=logicsampling then curvalue := lwtc[target,i]/likelihoodcount;
     flush(outfile);
     if curvalue > maxvalue then
         begin
             maxvalue := curvalue;
             curid := i;
         end;
   end;
  twriteint(outfile,round(cats[target].value[curid]));
 tpwritereal(outfile,maxvalue,5,4,false); 
end;

(****************************************************************************)
(* END of OUTPUT section                                                    *)
(****************************************************************************)

(****************************************************************************)
(* TEST section                                                             *)
(****************************************************************************)

procedure testDOlikelihoodweighting;
   var
	converged:boolean;
	sample:vertarray;
	time,weight,error:real;
	starttime:real;
	outfile:text;
	filename:inbuf;
   begin
	eraserealtallycard(lwtc);
	twriteeoln(output);
	filename := 'likelihood.test';
	twrite(output,'Filename for testing likelihood weighting:|');
	twriteln(output,filename);
	rewritefile(outfile,filename);
	starttime:=tgettime;
	time:=starttime;
	converged:=false;
	samplecount:=0;
	likelihoodcount:=0;
	while (not converged) do
	   begin
		getlikelihoodsample(sample,evidenceset,evidencelist);	{ DO likelihood update }
		samplecount := samplecount + 1;
		weight:=getlikelihoodscore(sample,evidenceset);
		likelihoodcount:=likelihoodcount + weight;
		realtally(sample,lwtc,weight);
		time:=tgettime;
		if ((samplecount mod 40)= 0)
		then
		   begin
			fillrealcard(lwtc,likelihoodcount,lwfreqcard);
			error:=measurederror(exactfreqcard,lwfreqcard);
			tpwritereal(outfile,time-starttime,10,3,false);
			twrite(outfile,' |');
			tpwritereal(outfile,error,10,8,false);
			twrite(outfile,' |');
			twriteint(outfile,samplecount );
			twriteln(outfile,' |');
		   end;
		converged:=(samplecount>10000);
	   end;
   end;




procedure testDOlogicsampling;
   var
	converged:boolean;
	sample:vertarray;
	error,time:real;
	starttime:real;
	outfile:text;
	filename:inbuf;
   begin
	hitcount:=0;
	eraseinttallycard(lstc);
	twriteeoln(output);
	filename := 'logicsamp.test';
	twrite(output,'Filename for testing logic sampling:|');
	twriteln(output,filename);
	rewritefile(outfile,filename);
	starttime:=tgettime;
	time:=starttime;
	converged:=false;
	samplecount:=0;
	while (not converged) do
	   begin
		{getgoodlogicsample(sample);}
		getlogicsample(sample);
		samplecount := samplecount + 1;
		if (matchsampletoevidence(evidencelist,sample))
		then
		   begin    {a hit for logic sampling }
			hitcount := hitcount + 1;
			inttally(sample,lstc);
		   end;

		time:=tgettime;
		if ((samplecount mod 40)= 0)
		then
		   begin
			fillcard(lstc,hitcount,lsfreqcard);
			error:=measurederror(exactfreqcard,lsfreqcard);
			tpwritereal(outfile,time-starttime,10,3,false);
			twrite(outfile,' |');
			tpwritereal(outfile,error,10,8,false);
			twrite(outfile,' |');
			twriteint(outfile,samplecount );
			twriteln(outfile,' |');
		   end;
		converged:=(samplecount>10000);
	   end;
   end;


procedure getgoodlogicsample(var sample:vertarray);
   label 999;
   var
	frontptr: eqorder; 
	effect: integer;
   begin
999:	frontptr:= varorder;
	while (frontptr^.next <> nil) do
	   begin
		effect := frontptr^.varbl;
		getdiscreteval(effect,sample,cats);  { get value for effect }
		frontptr:=frontptr^.next;
		if effect in evidenceset then
			if sample[effect]<>valueof(effect,evidencelist) then goto 999;
	   end;
   end;


procedure testDOmodlogicsampling;
   var
	converged:boolean;
	sample:vertarray;
	error,time:real;
	starttime:real;
	outfile:text;
	filename:inbuf;
   begin
	eraseinttallycard(lstc);
	twriteeoln(output);
	filename := 'modlogicsamp.test';
	twrite(output,'Filename for testing logic sampling:|');
	twriteln(output,filename);
	rewritefile(outfile,filename);
	starttime:=tgettime;
	time:=starttime;
	converged:=false;
	samplecount:=0;
	hitcount:=0;
	while (not converged) do
	   begin
		getgoodlogicsample(sample);
		samplecount := samplecount + 1;
		hitcount := hitcount + 1;
		inttally(sample,lstc);

		time:=tgettime;
		if ((samplecount mod 40)= 0)
		then
		   begin
			fillcard(lstc,hitcount,lsfreqcard);
			error:=measurederror(exactfreqcard,lsfreqcard);
			tpwritereal(outfile,time-starttime,10,3,false);
			twrite(outfile,' |');
			tpwritereal(outfile,error,10,8,false);
			twrite(outfile,' |');
			twriteint(outfile,samplecount );
			twriteln(outfile,' |');
		   end;
		converged:=(samplecount>10000);
	   end;
   end;


procedure testevidence;

   begin
	initevidence(evidencelist);

	{getevidence(evidencelist,evidenceset);
	printevidence(evidencelist,output);
	removeevidence(evidencelist,evidenceset);
	getevidence(evidencelist,evidenceset);
	printevidence(evidencelist,output);
	twriteln(output,'Deleteing all evidence|');
	purgeevidence(evidencelist,evidenceset);
	printevidence(evidencelist,output);}

	getevidence(evidencelist,evidenceset);
	printevidence(evidencelist,output);
   end;

(****************************************************************************)
(* END of TEST section                                                      *)
(****************************************************************************)

(****************************************************************************)
(* MAIN section                                                             *)
(****************************************************************************)

procedure addev; 
begin
   getevidence(evidencelist,evidenceset);
end;

procedure removeev;
begin 
  removeevidence(evidencelist,evidenceset);
end;

procedure removeallevidence;
begin
   purgeevidence(evidencelist,evidenceset);
end;

procedure initev;
begin
   initevidence(evidencelist);
end;

procedure updatebayesnetwork(* printit *);
var foundtheo: boolean; i: integer; curpos: eqorder;
begin
   foundtheo := true;
   if (varorder <> nil) then curpos := varorder^.next else curpos := varorder;
   while (curpos <> nil) and (not foundtheo) do
      begin
        if istheoretical(vernames[curpos^.varbl]) then
               myerror('Cannot update Bayes Network with latent variables.|',
                        fatal,false);
        curpos := curpos^.next;
      end;
    if not aaabort then
        begin
        	updatebayes;
                if printit then printbayesoutput(ofilvar)
(*                else printbayesoutput1(ofilvar); *)
        end;
end;

procedure printev(* var tofile: text *);
begin
   printevidence(evidencelist,output);
end;

function hadevidence; (* : boolean; *)
begin
   if evidencelist = nil then hadevidence := false
   else hadevidence := evidencelist^.next <> nil;
end;

(****************************************************************************)
(* END of MAIN section                                                      *)
(****************************************************************************)

(*-----------#pc-version-----------------------------------------------------*)
(*end.*)
(*-----------#pc-version-----------------------------------------------------*)

(* procedure mse;
var file1,file2: text; file1name, file2name: inbuf; changed: boolean;
    i,x,y,z,yest,zeroval,oneval,conditioned,cond1,cond2: integer; mse: real; 
begin
    initev;
    evidenceset := [];
    getfile(file1,'Sample data: ',know,file1name,changed);
    repeat
        readfileline(file1,buf);
    until (buf[1] = '/') and (buf[2] = 'R');
    readfileline(file1,buf);
    readfileline(file1,buf);
    mse := 0;
    repeat
        readfileline(file1,buf);
        x := round(readnum(false,fatal));
        y := round(readnum(false,fatal));
        z := round(readnum(false,fatal));
        conditioned := listposition('lc      ',false,vernames,changed);
        cond1 := listposition('ncig    ',false,vernames,changed);
        cond2 := listposition('hc      ',false,vernames,changed);
        evidenceset := [cond1];
        addevidence(evidencelist,evidenceset,vernames[1],1,x);
        evidenceset := [1..3] - [conditioned];
        addevidence(evidencelist,evidenceset,vernames[cond1],cond1,x);
        addevidence(evidencelist,evidenceset,vernames[cond2],cond2,z);
        updatebayes;
        if cats[conditioned].value[1] = 0 then zeroval := 1 else zeroval := 2;
        if zeroval = 1 then oneval := 2 else oneval := 1;
        if exactfreqcard[conditioned,zeroval] >= exactfreqcard[conditioned,oneval] then yest := 0
        else yest := 1;
        mse := mse + (yest - y) * (yest - y);
      until eof(file1);
    twrite(ofilvar,'mse = |');
    twritereal(ofilvar,mse);
 end;
*)

procedure readheader1(var file1: text; var map: parray);
var leaveloop,wasinserted,firsttime: boolean; name: vertexnametype; i,k: vertextype; 
   begin
     leaveloop := false;
     firsttime := true;
     i := 0;
     k := 1;
     while (not eof(file1)) and (not leaveloop) do 
      while (not leaveloop) do
       begin
         readwhitespace;
         if (not myeoln) then
          if alphabetic(buf[curpos]) then
           begin
             firsttime := false;
             readwhitespace;
             if (not myeoln) then
	       begin
                 readname(name,fatal);
                 if name[1] in ['A'..'Z'] then
   	           myerror('No theoreticals allowed in covariances.|  ',
                           fatal,true)
                 else i := i + 1;
                 map[k] := listposition(name,true,vernames,wasinserted);
                 k := k + 1;
	       end;
           end
          else 
              begin
               leaveloop := true;
               if firsttime then errormess('List of variables required here. |',fatal,true);
              end
         else readfileline(file1,buf);
        end;
end;

procedure updatedata;
var j: vertextype; val,targetval: real; i,temp1,temp2,temp3,temp4,k: integer;
    wasinserted: boolean; targetname: vertexnametype; file1: text; file1name: inbuf; changed: boolean;
    firstvalue,curvalue,thevalue: real; curid: vertextype; map: parray;
begin
    temp1 := 0; temp2 := 0; temp3 := 0; temp4 := 0;
    twrite(output,'Target Variable: |');
    readttyline(buf);  
    readname(targetname,ignoreall);
    target := listposition(targetname,false,vernames,wasinserted);
    getfile(file1,'Sample data: ',know,file1name,changed);
    repeat
        readfileline(file1,buf);
    until (buf[1] = '/') and (buf[2] = 'R');
    readfileline(file1,buf);
    sampsize := readnum(false,fatal);
    readfileline(file1,buf);
    readheader1(file1,map);
    for i := 1 to round(sampsize) do
      begin
        purgeevidence(evidencelist,evidenceset);
        initev;
        evidenceset := [];
        for j := 1 to max do
          if (map[j] in graphvertices) and (map[j] <> target) then evidenceset := evidenceset + [map[j]];
        for j := 1 to max do
          if (map[j] in graphvertices) and (map[j] <> target) then
            begin
              val := round(readnum(false,fatal));
              addevidence(evidencelist,evidenceset,vernames[map[j]],map[j],val);
            end
          else targetval := round(readnum(false,fatal));
        updatebayesnetwork(false);
        twrite(ofilvar,'    |');
        twriteint(ofilvar,round(targetval));
        for k := 1 to cats[target].num do
           begin
             if updatemethod=analytic then curvalue := exactfreqcard[target,k]
             else if updatemethod=likelihoodweighting then curvalue := lwtc[target,k]/likelihoodcount
             else if updatemethod=logicsampling then curvalue := lwtc[target,k]/likelihoodcount;
             flush(ofilvar);
             if round(cats[target].value[k]) = 1 then
                   thevalue := curvalue;
         end;
        twrite(ofilvar,'    |');
        tpwritereal(ofilvar,thevalue,5,4,false);
        twriteeoln(ofilvar);
        if (thevalue <= 0.5) and (targetval = 0) then temp1 := temp1 + 1;
        if (thevalue <= 0.5) and (targetval = 1) then temp2 := temp2 + 1;
        if (thevalue > 0.5) and (targetval = 0) then temp3 := temp3 + 1;
        if (thevalue > 0.5) and (targetval = 1) then temp4 := temp4 + 1; 
        readfileline(file1,buf);
      end;
    twriteint(ofilvar,temp1); twriteeoln(ofilvar);
    twriteint(ofilvar,temp2); twriteeoln(ofilvar);
    twriteint(ofilvar,temp3); twriteeoln(ofilvar);
    twriteint(ofilvar,temp4); twriteeoln(ofilvar);
    twrite(ofilvar,' final |');
    twriteint(ofilvar,temp1 + temp4); twriteeoln(ofilvar);
 end;

procedure domarg(marginal1,marginal2: vertextype; numUnknown,size: integer;
                     unknownarray: varray);
   var
	i:vertextype;
	j:integer;
        index1,index2: integer;
	sample:vertarray;
	probval:real;
   begin
        for i := 0 to maxcategories do
          begin
             marg1[i] := 0;
             marg2[i] := 0;
             for j := 0 to maxcategories do marg[i,j] := 0;
          end;
        initupdate;
	{	fill in statespace  }
	condprob:=0;
	for j:= 0 to (size - 1) do
	   begin
		decode(j,size,sample,numUnknown,unknownarray,cats);
                probval := statespace^.space[j];
                marg[round(sample[marginal1]),round(sample[marginal2])] := 
                    marg[round(sample[marginal1]),round(sample[marginal2])] + probval;
                marg1[round(sample[marginal1])] := marg1[round(sample[marginal1])] + probval;
                marg2[round(sample[marginal2])] := marg2[round(sample[marginal2])] + probval;
	   end;
   end;

function edgestrength(tail,head: vertextype; numUnknown,size: integer;
                     unknownarray: varray): real;
var i,j: vertextype; strength: real;
begin
  domarg(tail,head,numUnknown,size,unknownarray);
  strength := 0;
  for i := 0 to cats[tail].num - 1 do
     for j := 0 to cats[head].num - 1 do
       if (marg1[i] <> 0) and (marg2[j] <> 0) and (marg[i,j] <> 0) then
           strength := strength + 
             marg[i,j] * ln(marg[i,j] / (marg1[i] * marg2[j]));
  edgestrength := strength;
end;

procedure getedgestrengths;
var i,j: integer; sample:vertarray; 
	numUnknown,size:integer;
	unknownarray:varray;
	probval:real;
begin
   initevidence(evidencelist);
   size := getstatespace;
   statespace^.size:=size;
   evidenceset := [];
   numUnknown:=sizeofset(graphvertices)-sizeofset(evidenceset);
   j:=0;

	{set up unknownarray}
   for i:= 1 to origmax do
     if ((not (i in evidenceset)) and (i in graphvertices)) then
	   begin
		j:=j+1;
		unknownarray[j]:=i;
           end;

	{	fill in statespace  }

   for j := 0 to (size - 1) do
     begin
       decode(j,size,sample,numUnknown,unknownarray,cats);
       probval:=getjointprob(sample);
       statespace^.space[j]:=probval;
     end;
   for i := 1 to max do
      for j := 1 to max do
        if adjacency^[i,j] <> notthere then
             parm^[i,j] := edgestrength(i,j,numUnknown,size,unknownarray);
   dispose(evidencelist);
end;

{if restore comment out parameters
procedure getcrossentropy(var crossent: real);
var     
        i,j,k,l,multiplier : integer;
        sample,sample1	   : vertarray;
        index		   : real;
	numUnknown,size	   : integer;
	unknownarray	   : varray;
	probval		   : real;
        sspace1,sspace2	   : ^statespacetype;
        aneqs		   : eqorder;
begin
   new(sspace1);
   new(sspace2);
   initevidence(evidencelist);
   getknowledge;   
   formtree(adjacency,cats);
   new(aneqs);
   formeqsorder(aneqs,adjacency);
   for i := 1 to max do
      for j:= 1 to maxcategories do 
         cats[i].value[j] := varcatvalue[i,j];
   fillouttree(cats,MLexternal,1);
   size := getstatespace;
   sspace1^.size:=size;
   sspace2^.size:=size;
   for i:=0 to (size - 1) do
   begin
      sspace2^.space[i]:=0;
      sspace1^.space[i]:=0;
   end;   
   evidenceset := [];
   numUnknown:=sizeofset(graphvertices)-sizeofset(evidenceset);
   j:=0;
   for i:= 1 to origmax do
     if ((not (i in evidenceset)) and (i in graphvertices)) then
	   begin
		j:=j+1;
		unknownarray[j]:=i;
	   end;
   for j := 0 to (size - 1) do
     begin
       decode(j,size,sample,numUnknown,unknownarray,cats);
       index := 0;
       for k := 1 to max do 
           begin
             multiplier := 1;
             for l := k+1 to max do multiplier := multiplier * cats[l].num;
             index := index + multiplier *  (getnumcat(k,sample[k],cats)-1);
           end;
       probval:=getjointprob(sample);
       sspace1^.space[round(index)]:=probval;
     end;
   getknowledge; 
   for j := 0 to (size - 1) do
     begin
       decode(j,size,sample,numUnknown,unknownarray,cats);
       for k := 1 to max do 
           sample1[k] := varcatvalue[k,round(sample[k]+1)];
       index := 0;
       for k := 1 to max do 
           begin
             multiplier := 1;
             for l := k+1 to max do multiplier := multiplier * cats[l].num;
             index := index + multiplier *  (getnumcat(k,sample1[k],cats)-1);
           end;
       probval:=getjointprob(sample);
       sspace2^.space[round(index)]:=probval;
     end;
   crossent := 0;
   for j := 0 to size - 1 do
     begin
      if (sspace2^.space[j] <> 0) and (sspace1^.space[j] <> 0) then
        crossent := crossent + (sspace2^.space[j] * ln(sspace2^.space[j] /
                                                       sspace1^.space[j]));
     end;
   dispose(sspace1);
   dispose(sspace2);
   disposeeqsorder(aneqs);
   twritereal(ofilvar,crossent);
   twriteeoln(ofilvar);
end;

}


procedure getcrossentropy(* var crossent1,crossent3: real *);
(* This assumes that variables are in same order in mod file and dat file. *)
var     
        localvarcat: catnumarray;
        i,j,k,l,multiplier : integer;
        curvar: vertextype;
sample,sample1			   : vertarray;
index			   : real;
	numUnknown,size	   : integer;
	unknownarray	   : varray;
	probval		   : real;
        sspace1,                       (* estimated from graph *)
        sspace3     : ^statespacetype; (* population distribution *)
        aneqs		   : eqorder;
   sum1,sum2,sum3,sum4		   : real;
   localcats: categoryarray;
begin
   l := 6;
   for curvar := 1 to max do
      if curvar in graphvertices then
         begin
           write(output,'Number of categories for ');
           rswritename(curvar,output,l);
           writeln(output);
           readln(input,cats[curvar].num);
         end;
   new(sspace1);
   new(sspace3);
   initevidence(evidencelist);
   write(output,'get the data ');
   getknowledge;   (* get the data file *)
   write(output,'get the correct graph ');
   getknowledge;   (* get the graph file *)
   for i := 1 to max do
      for j:= 1 to maxcategories do 
        begin
         cats[i].value[j] := varcatvalue[i,j];
         localcats[i].value[j] := cats[i].value[j];
         localvarcat[i,j] := varcatvalue[i,j];
        end;
   formtree(adjacency,cats);
   new(aneqs);
   formeqsorder(aneqs,adjacency);
   for i := 1 to max do 
      begin
         cats[i].num := maxval[i];
         localcats[i].num := cats[i].num;
     end;
   fillouttree(cats,MLexternal,1);
(*   writediscretemodel(cats,aneqs,adjacency,ofilvar,false); *)
   size := getstatespace;
   sspace1^.size:=size;
   for i:=0 to (size - 1) do
   begin
      statespace^.space[i]:=0;
      sspace1^.space[i]:=0;
      sspace3^.space[i] := 0;
   end;   
   evidenceset := [];
   numUnknown:=sizeofset(graphvertices)-sizeofset(evidenceset);
   j:=0;

	{set up unknownarray}
   for i:= 1 to origmax do
     if ((not (i in evidenceset)) and (i in graphvertices)) then
	   begin
		j:=j+1;
		unknownarray[j]:=i;
	   end;

	{	fill in statespace  }
   for i:=1 to round(sampsize) do    (* statespace is sample distribution *)
   begin
      for j:=1 to max do
	 if (j in graphvertices) then
	    sample[j] := getbigdata(i,j) - 1;
      index:=0;
      for k := 1 to max do 
      begin
	 multiplier := 1;
	 for l := k+1 to max do multiplier := multiplier * cats[l].num;
	 index := index + multiplier * sample[k];
      end;
      statespace^.space[round(index)]:= statespace^.space[round(index)] + 1;
   end;
   for i:=0 to size -1 do
      statespace^.space[i]:= statespace^.space[i]/round(sampsize);      
   numz := 0;
   for j := 0 to (size - 1) do
     begin
       decode(j,size,sample,numUnknown,unknownarray,cats);
       index := 0;
       for k := 1 to max do 
           begin
             multiplier := 1;
             for l := k+1 to max do multiplier := multiplier * cats[l].num;
             index := index + multiplier *  (getnumcat(k,sample[k],cats)-1);
           end;
       probval:=getjointprob(sample);
       if probval = 0 then
          numz := numz + 1;
       sspace1^.space[round(index)]:=probval;
     end;
   write('get the population distribution ');
   getknowledge;
   for j := 0 to (size - 1) do
     begin
       decode(j,size,sample1,numUnknown,unknownarray,localcats);
(*       for k := 1 to max do 
           sample1[k] := varcatvalue[k,round(sample1[k]+1)]; *)
       index := 0;
       for k := 1 to max do 
           begin
             multiplier := 1;
             for l := k+1 to max do multiplier := multiplier * localcats[l].num;
             index := index + multiplier *  (getnumcat(k,sample1[k],localcats)-1);
           end;
       probval:=getjointprob(sample1);
       sspace3^.space[round(index)]:=probval;
     end;
   crossent1 := 0;
   crossent3 := 0;
   sum1 := 0;
   sum2 := 0;
   sum4 := 0;
   for j := 0 to size - 1 do
     begin
      sum1 := sum1 + statespace^.space[j];
      sum2 := sum2 + sspace1^.space[j];
      sum4 := sum4 + sspace3^.space[j];
      if (sspace3^.space[j] <> 0) and (sspace1^.space[j] = 0) then
         crossent1 := -1;
      if crossent1 <> -1 then
         crossent1 := crossent1 + (sspace3^.space[j] * ln(sspace3^.space[j] /
                                                       sspace1^.space[j]));
      if (sspace3^.space[j] <> 0) and (statespace^.space[j] = 0) then 
         crossent3 := -1;
      if crossent3 <> -1 then
         crossent3 := crossent3 + (sspace3^.space[j] * ln(sspace3^.space[j] /
                                                       statespace^.space[j]));
     end;
   dispose(sspace1);
   dispose(sspace3);
   disposeeqsorder(aneqs);
end;


procedure mse1(a: integer);
var file1,file2: text; file1name, file2name: inbuf; changed: boolean;
    i,x,y,z,yest,zeroval,oneval,conditioned,cond1,cond2: integer; mse: real; 
begin
    initev;
    evidenceset := [];
    getfile(file1,'Sample data: ',know,file1name,changed);
    repeat
        readfileline(file1,buf);
    until (buf[1] = '/') and (buf[2] = 'R');
    readfileline(file1,buf);
    readfileline(file1,buf);
    mse := 0;
    repeat
        readfileline(file1,buf);
        x := round(readnum(false,fatal));
        y := round(readnum(false,fatal));
        z := round(readnum(false,fatal));
        conditioned := listposition('lc      ',false,vernames,changed);
        cond1 := listposition('ncig    ',false,vernames,changed);
        cond2 := listposition('hc      ',false,vernames,changed);
        evidenceset := [cond1];
        addevidence(evidencelist,evidenceset,vernames[1],1,x);
        evidenceset := [1..3] - [conditioned];
        addevidence(evidencelist,evidenceset,vernames[cond1],cond1,x);
        addevidence(evidencelist,evidenceset,vernames[cond2],cond2,z);
        updatebayes;
        if cats[conditioned].value[1] = 0 then zeroval := 1 else zeroval := 2;
        if zeroval = 1 then oneval := 2 else oneval := 1;
        if exactfreqcard[conditioned,zeroval] >= exactfreqcard[conditioned,oneval] then yest := 0
        else yest := 1;
        mse := mse + (yest - y) * (yest - y);
      until eof(file1);
    twrite(ofilvar,'mse = |');
    twritereal(ofilvar,mse);
 end;
